"""
Tests for the CosConnection
"""

# Third Party
import pytest

# Local
from oper8.exceptions import PreconditionError
from oper8.test_helpers.helpers import MockDeployManager, setup_session
from oper8.test_helpers.oper8x_helpers import set_secret_data
from oper8.x.datastores.cos import CosConnection, CosFactory
from oper8.x.utils import common, constants

## Helpers #####################################################################

sample_base_kwargs = {
    "hostname": "foo.bar.com",
    "port": 1234,
    "bucket_name": "test-bucket",
}

sample_auth_kwargs = {
    "auth_secret_name": "test-auth-secret",
    "auth_secret_access_key_field": "accessKey",
    "auth_secret_secret_key_field": "secretKey",
}

sample_tls_kwargs = {
    "tls_secret_name": "test-tls-secret",
    "tls_secret_cert_field": "ca.pem",
}

sample_uri_kwargs = {
    "uri_secret_name": "test-uri-secret",
    "uri_secret_hostname_field": "hostname",
    "uri_secret_port_field": "port",
    "uri_secret_bucketname_field": "bucketname",
}

sample_auth_secret_kwargs = {
    "access_key": "accessMe",
    "secret_key": "IAmVerySecret",
}

sample_tls_secret_kwargs = {
    "tls_cert": "No cert to be found",
}

sample_uri_secret_kwargs = {
    "hostname": "fake-hostname",
    "port": "12345",
    "bucketname": "fake-bucketname",
}

auth_secret_data = {
    sample_auth_kwargs["auth_secret_access_key_field"]: common.b64_secret(
        sample_auth_secret_kwargs["access_key"]
    ),
    sample_auth_kwargs["auth_secret_secret_key_field"]: common.b64_secret(
        sample_auth_secret_kwargs["secret_key"]
    ),
}

tls_secret_data = {
    sample_tls_kwargs["tls_secret_cert_field"]: common.b64_secret(
        sample_tls_secret_kwargs["tls_cert"]
    ),
}

uri_secret_data = {
    sample_uri_kwargs["uri_secret_hostname_field"]: common.b64_secret(
        sample_uri_secret_kwargs["hostname"]
    ),
    sample_uri_kwargs["uri_secret_port_field"]: common.b64_secret(
        sample_uri_secret_kwargs["port"]
    ),
    sample_uri_kwargs["uri_secret_bucketname_field"]: common.b64_secret(
        sample_uri_secret_kwargs["bucketname"]
    ),
}


def set_auth_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_auth_kwargs["auth_secret_name"],
        data=auth_secret_data,
    )


def set_tls_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_tls_kwargs["tls_secret_name"],
        data=tls_secret_data,
    )


def set_uri_secret_data(session):
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_uri_kwargs["uri_secret_name"],
        data=uri_secret_data,
    )


def set_cos_secrets(session):
    set_auth_secret_data(session)
    set_tls_secret_data(session)
    set_uri_secret_data(session)


COS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES = {
    constants.SPEC_DATASTORES: {
        CosFactory.DATASTORE_TYPE: {
            constants.SPEC_DATASTORE_CONNECTION: {
                "uri_secret": sample_uri_kwargs["uri_secret_name"],
                "uri_secret_hostname_field": sample_uri_kwargs[
                    "uri_secret_hostname_field"
                ],
                "uri_secret_port_field": sample_uri_kwargs["uri_secret_port_field"],
                "uri_secret_bucketname_field": sample_uri_kwargs[
                    "uri_secret_bucketname_field"
                ],
                "auth_secret_name": sample_auth_kwargs["auth_secret_name"],
                "auth_secret_access_key_field": sample_auth_kwargs[
                    "auth_secret_access_key_field"
                ],
                "auth_secret_secret_key_field": sample_auth_kwargs[
                    "auth_secret_secret_key_field"
                ],
                "tls_secret_name": sample_tls_kwargs["tls_secret_name"],
                "tls_secret_cert_field": sample_tls_kwargs["tls_secret_cert_field"],
            }
        }
    }
}


## Tests #######################################################################

#################
## Constructor ##
#################


def test_construct_no_secrets():
    """Test that constructing a connection without secrets works as expected and
    does not have the secret values pre-populated
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    for key, val in sample_base_kwargs.items():
        assert getattr(conn, key) == val
    for key, val in sample_auth_kwargs.items():
        assert getattr(conn, key) == val
    for key, val in sample_tls_kwargs.items():
        assert getattr(conn, key) == val
    assert conn._access_key is None
    assert conn._secret_key is None
    assert conn._tls_cert is None


def test_construct_with_secrets():
    """Test that constructing a connection with secrets works as expected and
    has the secret values pre-populated
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
        **sample_auth_secret_kwargs,
        **sample_tls_secret_kwargs,
    )
    for key, val in sample_base_kwargs.items():
        assert getattr(conn, key) == val
    for key, val in sample_auth_kwargs.items():
        assert getattr(conn, key) == val
    for key, val in sample_tls_kwargs.items():
        assert getattr(conn, key) == val
    assert conn._access_key == sample_auth_secret_kwargs["access_key"]
    assert conn._secret_key == sample_auth_secret_kwargs["secret_key"]
    assert conn._tls_cert == sample_tls_secret_kwargs["tls_cert"]


def test_construct_no_tls():
    """Test that constructing a connection without tls values works as expected
    and correctly determines that tls is disabled
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
    )
    assert not conn.tls_enabled
    assert conn.schema == "http"


def test_construct_with_tls():
    """Test that constructing a connection with tls values works as expected
    and correctly determines that tls is enabled
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    assert conn.tls_enabled
    assert conn.schema == "https"


def test_construct_tls_args_inconsistent():
    """Test that constructing with inconsistent tls args raises an exception"""
    session = setup_session()
    with pytest.raises(AssertionError):
        CosConnection(
            session,
            tls_secret_name="not-going-to-give-cert-field",
            **sample_base_kwargs,
            **sample_auth_kwargs,
        )
    with pytest.raises(AssertionError):
        CosConnection(
            session,
            tls_secret_cert_field="not-going-to-give-secret-name",
            **sample_base_kwargs,
            **sample_auth_kwargs,
        )
    with pytest.raises(AssertionError):
        CosConnection(
            session,
            tls_cert="not-going-to-give-secret-info",
            **sample_base_kwargs,
            **sample_auth_kwargs,
        )


################
## Properties ##
################


def test_all_properties():
    """Test that all properties reflect the constructed values and that they are
    read-only
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    for property in [
        "schema",
        "hostname",
        "port",
        "endpoint",
        "bucket_name",
        "auth_secret_name",
        "auth_secret_access_key_field",
        "auth_secret_secret_key_field",
        "tls_secret_name",
        "tls_secret_cert_field",
        "tls_enabled",
    ]:
        assert hasattr(conn, property)
        with pytest.raises(AttributeError):
            setattr(conn, property, "foobar")


###################
## Serialization ##
###################


def test_to_dict_all_required_fields():
    """Make sure that all expected fields are present. This is a bit silly since
    it's just a copy-paste of the list of keys, but it's less likely that we
    screw that up in two places than in one!
    """
    session = setup_session()
    conn_dict = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    ).to_dict()
    for key in [
        "hostname",
        "port",
        "bucket_name",
        "auth_secret_name",
        "auth_secret_access_key_field",
        "auth_secret_secret_key_field",
        "tls_secret_name",
        "tls_secret_cert_field",
    ]:
        assert key in conn_dict


def test_to_dict_snake_case_keys():
    """Make sure that the keys returned by to_dict maintain their snake casing."""
    session = setup_session()
    conn_dict = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    ).to_dict()
    assert common.camelcase_to_snake_case(conn_dict) == conn_dict


def test_from_dict_camel_case_values_preserved():
    """Make sure that values in the config dict which are camelCase are
    preserved during from_dict
    """
    session = setup_session()
    all_kwargs = dict(
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    camel_dict = common.snake_case_to_camelcase(all_kwargs)
    conn = CosConnection.from_dict(session, camel_dict)
    for key, val in all_kwargs.items():
        assert getattr(conn, key) == val


def test_from_dict_missing_keys():
    """Make sure that calling from_dict with missing keys raises an error"""
    session = setup_session()
    all_kwargs = dict(
        **sample_base_kwargs,
        **sample_tls_kwargs,
    )
    camel_dict = common.snake_case_to_camelcase(all_kwargs)
    with pytest.raises(ValueError):
        CosConnection.from_dict(session, camel_dict)


def test_from_dict_extra_keys():
    """Make sure that calling from_dict with extra keys ignores them"""
    session = setup_session()
    all_kwargs = dict(
        foo="bar",
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    camel_dict = common.snake_case_to_camelcase(all_kwargs)
    CosConnection.from_dict(session, camel_dict)


###################
## get_auth_keys ##
###################


def test_get_auth_keys_secret_found():
    """Test that get_auth_keys returns the secret values when the secret is
    found and the expected keys are present
    """
    session = setup_session(
        deploy_manager=MockDeployManager(),
    )
    set_auth_secret_data(session)
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    access_key, secret_key = conn.get_auth_keys()
    assert access_key == sample_auth_secret_kwargs["access_key"]
    assert secret_key == sample_auth_secret_kwargs["secret_key"]


def test_get_auth_keys_missing_keys():
    """Test that get_auth_keys returns no values if the secret is found, but the
    expected keys are not found
    """
    session = setup_session(
        deploy_manager=MockDeployManager(),
    )
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_auth_kwargs["auth_secret_name"],
        data={
            sample_auth_kwargs["auth_secret_access_key_field"]: common.b64_secret(
                sample_auth_secret_kwargs["access_key"]
            ),
        },
    )
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    access_key, secret_key = conn.get_auth_keys()
    assert access_key is None
    assert secret_key is None


def test_get_auth_keys_secret_not_found():
    """Test that get_auth_keys returns no values if the secret is not found"""
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    access_key, secret_key = conn.get_auth_keys()
    assert access_key is None
    assert secret_key is None


def test_get_auth_keys_prepopulated():
    """Test that get_auth_keys doesn't look in the cluster if the values are
    given at construct time
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
        **sample_auth_secret_kwargs,
    )
    access_key, secret_key = conn.get_auth_keys()
    assert access_key == sample_auth_secret_kwargs["access_key"]
    assert secret_key == sample_auth_secret_kwargs["secret_key"]


##################
## get_tls_cert ##
##################


def test_get_tls_cert_secret_found():
    """Test that get_tls_cert returns the secret value when the secret is
    found and the expected key is present
    """
    session = setup_session(
        deploy_manager=MockDeployManager(),
    )
    set_tls_secret_data(session)
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    tls_cert = conn.get_tls_cert()
    assert tls_cert == sample_tls_secret_kwargs["tls_cert"]


def test_get_tls_cert_missing_key():
    """Test that get_tls_cert returns no value if the secret is found, but the
    expected key is not found
    """
    session = setup_session(
        deploy_manager=MockDeployManager(),
    )
    set_secret_data(
        session=session,
        scoped_name=False,
        name=sample_tls_kwargs["tls_secret_name"],
        data={},
    )
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    tls_cert = conn.get_tls_cert()
    assert tls_cert is None


def test_get_tls_cert_secret_not_found():
    """Test that get_tls_cert returns no value if the secret is not found"""
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    tls_cert = conn.get_tls_cert()
    assert tls_cert is None


def test_get_tls_cert_prepopulated():
    """Test that get_tls_cert doesn't look in the cluster if the value is given
    at construct time
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
        **sample_tls_secret_kwargs,
    )
    tls_cert = conn.get_tls_cert()
    assert tls_cert == sample_tls_secret_kwargs["tls_cert"]


def test_get_tls_cert_disabled():
    """Test that get_tls_cert reurns None if tls is disabled"""
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
    )
    tls_cert = conn.get_tls_cert()
    assert tls_cert is None


###########################
## get_connection_string ##
###########################


def test_get_connection_string_secrets_known():
    """Make sure the formatting is right in get_connection_string. This is also
    a bit of a sily test because it's just a copy-and-paste of the
    implementation, but two we'll call it an A/B test. :shrug:
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
        **sample_auth_secret_kwargs,
    )
    connection_string = conn.get_connection_string()
    assert (
        connection_string
        == "s3,endpoint={}://{}:{},accesskey={},secretkey={},bucketsuffix={}".format(
            conn.schema,
            conn.hostname,
            conn.port,
            sample_auth_secret_kwargs["access_key"],
            sample_auth_secret_kwargs["secret_key"],
            conn.bucket_name,
        )
    )


def test_get_connection_string_missing_secrets():
    """Make sure that trying to get the connection string when the secrets are
    not available raises a PreconditionError
    """
    session = setup_session()
    conn = CosConnection(
        session,
        **sample_base_kwargs,
        **sample_auth_kwargs,
        **sample_tls_kwargs,
    )
    with pytest.raises(PreconditionError):
        conn.get_connection_string()


##############
## provided ##
##############


def test_get_component_with_provided_connection():
    """Test that a provided connection passed in through the CR returns None on a get_component call"""
    override_deploy_configs = COS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES

    session = setup_session(
        deploy_config=override_deploy_configs,
        deploy_manager=MockDeployManager(),
    )
    set_cos_secrets(session)
    component = CosFactory.get_component(session)
    # Since the component is provided, we should be getting None back
    assert component is None


def test_get_connection_with_provided_connection():
    """Test that a provided connection passed in through the CR returns valid connection details on get_connection call"""
    override_deploy_configs = COS_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES
    session = setup_session(
        deploy_config=override_deploy_configs,
        deploy_manager=MockDeployManager(),
    )
    set_cos_secrets(session)
    connection = CosFactory.get_connection(session)
    assert connection.hostname == sample_uri_secret_kwargs["hostname"]
    # Need to convert port value to int since Factory converts string input of port to type int
    assert connection.port == int(sample_uri_secret_kwargs["port"])
    assert connection.bucket_name == sample_uri_secret_kwargs["bucketname"]



"""
Test the implementations of the default functions in Component
"""

# Standard
from unittest import mock
import os
import tempfile

# Third Party
import pytest
import yaml

# First Party
import alog

# Local
from oper8 import constants
from oper8.component import Component, ManagedObject
from oper8.constants import TEMPORARY_PATCHES_ANNOTATION_NAME
from oper8.deploy_manager import DeployMethod
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.deploy_manager.owner_references import _make_owner_reference
from oper8.patch import STRATEGIC_MERGE_PATCH
from oper8.test_helpers.helpers import (
    TEST_NAMESPACE,
    DummyNodeComponent,
    MockDeployManager,
    configure_logging,
    library_config,
    make_patch,
    setup_session,
)
from oper8.utils import merge_configs

configure_logging()
log = alog.use_channel("TEST")


def get_comp_type(name="dummy"):
    """Parametrization helper to get test both standard and legacy components.
    This function also wraps the output class type so that name class attributes
    are not polluted.
    """
    given_name = name

    class Wrapped(DummyNodeComponent):
        name = given_name

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    return Wrapped


################################################################################
## Tests #######################################################################
################################################################################

##################
## Construction ##
##################


def test_name_validation():
    """Validate that the name validation passes for a class with a valid
    name class attribute
    """

    class NamedComponent(Component):
        name = "foo"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def build_chart(self, session):
            pass

    class UnnamedComponent(Component):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def build_chart(self, session):
            pass

    session = setup_session()
    component = NamedComponent(session=session)
    assert component.name == NamedComponent.name
    with pytest.raises(NotImplementedError):
        UnnamedComponent(session=session)


def test_scope():
    """Validate that the scope given to the constructor gets wired into the
    component's hierarchy correctly
    """
    session = setup_session()
    comp = get_comp_type()(session=session)
    assert session.get_component(comp.name)


##################
## render_chart ##
##################


def test_managed_objects():
    """Make sure managed_objects matches the objects generated by the
    chart
    """
    session = setup_session()
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
            ("bat", {"kind": "Baz", "metadata": {"name": "bat"}}),
        ],
    )
    comp.render_chart(session)
    managed_objs = comp.managed_objects
    assert len(managed_objs) == 2
    assert managed_objs[0].kind == "Foo"
    assert managed_objs[1].kind == "Baz"


def test_apply_patches_ok():
    """Make sure applying patches modifies the managed objects as expected"""
    patch_name = "test"
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, {"dummy": {"bar": {"new": "value"}}}, patch_name
    )
    log.debug("Patch: %s", patch)
    dm = DryRunDeployManager(resources=[patch])
    session = setup_session(temporary_patches=[patch])
    bar = {"kind": "Foo", "metadata": {"name": "bar"}}
    bat = {"kind": "Bat", "metadata": {"name": "bat"}}
    with library_config(internal_name_annotation=False):
        comp = get_comp_type()(
            session=session, api_objects=[("bar", bar), ("bat", bat)]
        )
        comp.render_chart(session)
        objs = comp.to_config(session)
    assert len(objs) == 2
    expected_bar = merge_configs(
        bar,
        {
            "new": "value",
            "apiVersion": "v1",
            "metadata": {"namespace": TEST_NAMESPACE},
        },
    )
    assert objs[0] == expected_bar
    assert objs[1] == merge_configs(
        bat, {"apiVersion": "v1", "metadata": {"namespace": TEST_NAMESPACE}}
    )


def test_object_update():
    """Make sure in-place object updates work with update_object_definition"""
    session = setup_session()
    bar = {"kind": "Foo", "metadata": {"name": "bar"}}
    bat = {"kind": "Bat", "metadata": {"name": "bat"}}
    comp = get_comp_type()(session=session, api_objects=[("bar", bar), ("bat", bat)])

    def update_object_definition(session, internal_name, resource_definition):
        if internal_name == "dummy.bar":
            resource_definition["new"] = "value"
        return resource_definition

    comp.update_object_definition = update_object_definition
    with library_config(internal_name_annotation=False):
        comp.render_chart(session)
        objs = comp.to_config(session)
    assert len(objs) == 2
    expected_bar = merge_configs(
        bar,
        {
            "new": "value",
            "apiVersion": "v1",
            "metadata": {"namespace": TEST_NAMESPACE},
        },
    )
    assert objs[0] == expected_bar
    assert objs[1] == merge_configs(
        bat, {"apiVersion": "v1", "metadata": {"namespace": TEST_NAMESPACE}}
    )


def test_add_namespace():
    """Make sure that all internal components auto add the session.namespace to
    all resources
    """
    session = setup_session(namespace="unique")
    bar = {"kind": "Foo", "apiVersion": "v1", "metadata": {"name": "bar"}}

    class TestComponent(Component):
        name = "bar"

        def build_chart(self, session):
            self.add_resource("bar", bar)

    comp = TestComponent(session=session)

    with library_config(internal_name_annotation=True):
        comp.render_chart(session)
        objs = comp.to_config(session)
    assert len(objs) == 1
    assert objs[0].metadata.namespace == "unique"


def test_deploy_method():
    """Make sure that all internal components auto add the session.namespace to
    all resources
    """
    session = setup_session(namespace="unique")
    bar = {"kind": "Foo", "apiVersion": "v1", "metadata": {"name": "bar"}}
    bat = {"kind": "Foo", "apiVersion": "v1", "metadata": {"name": "bat"}}

    class TestComponent(Component):
        name = "bar"

        def build_chart(self, session):
            self.add_resource("replaced_bar", bar, deploy_method=DeployMethod.REPLACE)
            self.add_resource("replaced_bat", bat, deploy_method=DeployMethod.UPDATE)

    comp = TestComponent(session=session)

    arg_kwarg_list = []

    def append_call_args(*args, **kwargs):
        arg_kwarg_list.append([args, kwargs])
        return (True, True)

    with mock.patch.object(
        session.deploy_manager, "deploy", new=append_call_args
    ) as deployed:
        with library_config(internal_name_annotation=True):
            comp.render_chart(session)
            success = comp.deploy(session)

        assert arg_kwarg_list[0][1]["method"] == DeployMethod.REPLACE
        assert arg_kwarg_list[1][1]["method"] == DeployMethod.UPDATE

    assert success


def test_custom_resource_verify():
    """Make sure that a user  can override the verify function for a particular resource."""
    session = setup_session()
    bar = {"kind": "Foo", "apiVersion": "v1", "metadata": {"name": "bar"}}

    class TestSuccessVerify(Component):
        name = "bar"

        def build_chart(self, session):
            self.add_resource("bar", bar)

    class TestFailedVerify(Component):
        name = "bar_fail"

        def build_chart(self, session):
            self.add_resource("bar", bar, lambda resource: False)

    suc_comp = TestSuccessVerify(session=session)
    failed_comp = TestFailedVerify(session=session)

    with library_config(internal_name_annotation=True):
        suc_comp.render_chart(session)
        suc_comp.deploy(session)
        assert suc_comp.verify(session)

        failed_comp.render_chart(session)
        failed_comp.deploy(session)
        assert not failed_comp.verify(session)


def test_internal_name_annotation():
    """Make sure the internal name annotation is added to output resources if
    configured
    """
    session = setup_session()
    bar = {"kind": "Foo", "metadata": {"name": "bar"}}
    CompType = get_comp_type()
    comp = CompType(session=session, api_objects=[("bar", bar)])

    with library_config(internal_name_annotation=True):
        comp.render_chart(session)
        objs = comp.to_config(session)
    assert len(objs) == 1
    assert objs[0].metadata.annotations[
        constants.INTERNAL_NAME_ANNOTATION_NAME
    ] == ".".join([CompType.name, "bar"])


def test_cross_namespace_patch_annotation():
    """Make sure the temporary patch annotation is not added to resources of the same
    name in different namespaces
    """
    # Setup Session
    session = setup_session()
    foo = {
        "kind": "Foo",
        "apiVersion": "v1",
        "metadata": {
            "name": "foo",
            "namespace": session.namespace,
            "annotations": {
                TEMPORARY_PATCHES_ANNOTATION_NAME: "temp_patch",
            },
        },
    }
    bar = {
        "kind": "Foo",
        "api_version": "v1",
        "metadata": {"name": "foo", "namespace": "adifferentnamespace"},
    }

    # Deploy Foo and Remove apiVersion from dict
    session.deploy_manager.deploy([foo])
    del foo["apiVersion"]

    # Setup Component
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("foo", foo),
            ("bar", bar),
        ],
    )
    rendered = comp.to_dict(session)

    # Validate objects weren't incorrectly patched
    assert len(rendered) == 2
    assert rendered[0]["metadata"]["namespace"] == "adifferentnamespace"
    assert not rendered[0]["metadata"]["annotations"].get(
        TEMPORARY_PATCHES_ANNOTATION_NAME, None
    )
    assert rendered[1]["metadata"]["namespace"] == session.namespace
    assert (
        rendered[1]["metadata"]["annotations"][TEMPORARY_PATCHES_ANNOTATION_NAME]
        == "temp_patch"
    )


def test_dependency_order():
    """Make sure that objects are rendered in the correct order based on the
    declared dependencies in the component.
    """
    session = setup_session()
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
            ("bat", {"kind": "Baz", "metadata": {"name": "bat"}}),
        ],
        api_object_deps={"bar": ["bat"]},
    )
    comp.render_chart(session)
    managed_objs = comp.managed_objects
    assert len(managed_objs) == 2
    assert managed_objs[0].kind == "Baz"
    assert managed_objs[1].kind == "Foo"


def test_owner_references_external():
    """Make sure that external owner references are preserved"""
    owner_cr = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {
            "name": "owner",
            "namespace": "test",
            "uid": "unique",
        },
    }
    dm = DryRunDeployManager(
        resources=[
            {
                "kind": "Foo",
                "apiVersion": "v1",
                "metadata": {
                    "name": "foo",
                    "namespace": "test",
                    "uid": "new york",
                    "ownerReferences": [
                        {
                            "uid": "external",
                        }
                    ],
                },
            }
        ],
        owner_cr=owner_cr,
    )
    session = setup_session(deploy_manager=dm, namespace="test")
    comp = get_comp_type()(
        session=session,
        api_objects=[("foo", {"kind": "Foo", "api_version": "v1"})],
        full_cr=owner_cr,
    )
    comp.render_chart(session)
    comp.deploy(session)
    _, rendered = session.get_object_current_state(kind="Foo", name="foo")
    assert rendered is not None
    assert len(rendered["metadata"]["ownerReferences"]) == 2


def test_owner_references_no_duplicates():
    """Make sure that the current owner reference is not duplicated"""
    owner_cr = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {
            "name": "owner",
            "namespace": "test",
            "uid": "unique",
        },
    }
    dm = DryRunDeployManager(
        resources=[
            {
                "kind": "Foo",
                "apiVersion": "v1",
                "metadata": {
                    "name": "foo",
                    "namespace": "test",
                    "ownerReferences": [_make_owner_reference(owner_cr)],
                    "uid": "new york",
                },
            }
        ],
        owner_cr=owner_cr,
    )
    session = setup_session(deploy_manager=dm, full_cr=owner_cr, namespace="test")
    comp = get_comp_type()(
        session=session,
        api_objects=[("foo", {"kind": "Foo", "api_version": "v1"})],
        full_cr=owner_cr,
    )
    comp.render_chart(session)
    comp.deploy(session)
    _, rendered = session.get_object_current_state(kind="Foo", name="foo")
    assert rendered is not None
    assert len(rendered["metadata"]["ownerReferences"]) == 1


############
## deploy ##
############


def test_deploy_after_render():
    """Make sure that a deploy run after a render runs cleanly"""
    session = setup_session()
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
            ("bat", {"kind": "Baz", "metadata": {"name": "bat"}}),
        ],
    )
    comp.render_chart(session)
    assert comp.deploy(session)


def test_deploy_before_render():
    """Make sure that a deploy run before a render throws an appropriate
    exception
    """
    session = setup_session()
    comp = get_comp_type()(session=session)
    with pytest.raises(AssertionError):
        comp.deploy(session)


def test_deploy_failure():
    """Make sure that a failed deploy is captured successfully"""
    session = setup_session(deploy_manager=MockDeployManager(deploy_fail=True))
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
        ],
    )
    comp.render_chart(session)
    deploy_success = comp.deploy(session)
    assert not deploy_success


def test_deploy_raise():
    """Make sure that a thrown exception in a deploy will raise outside of
    the deploy function (to be handled by the rollout manager)
    """
    session = setup_session(deploy_manager=MockDeployManager(deploy_raise=True))
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
        ],
    )
    comp.render_chart(session)
    with pytest.raises(AssertionError):
        comp.deploy(session)


################
## Dependency ##
################


def test_add_dependency():
    session = setup_session(deploy_manager=MockDeployManager(deploy_raise=True))
    comp_a = get_comp_type("a")(session=session)
    comp_b = get_comp_type("b")(session=session)
    comp_a.add_dependency(session, comp_b)
    assert session.get_component_dependencies(comp_a) == [(comp_b, None)]


def test_add_multiple_dependency():
    session = setup_session(deploy_manager=MockDeployManager(deploy_raise=True))
    comp_a = get_comp_type("a")(session=session)
    comp_b = get_comp_type("b")(session=session)
    comp_c = get_comp_type("c")(session=session)
    comp_a.add_dependency(session, comp_b, comp_c)
    assert session.get_component_dependencies(comp_a) == [
        (comp_b, None),
        (comp_c, None),
    ]


def test_add_dependency_function():
    session = setup_session(deploy_manager=MockDeployManager(deploy_raise=True))
    comp_a = get_comp_type("a")(session=session)
    comp_b = get_comp_type("b")(session=session)
    verify_fn = lambda *_, **__: True
    comp_a.add_dependency(session, comp_b, verify_function=verify_fn)
    dep_list = session.get_component_dependencies(comp_a)
    assert len(dep_list) == 1
    assert dep_list[0][0] == comp_b
    assert dep_list[0][1].func == verify_fn


#############
## disable ##
#############


def test_disable_after_render():
    """Make sure disable runs successfully after rendering and correctly
    removes objects that are present.
    """
    deploy_manager = MockDeployManager(
        resources=[
            {
                "kind": "Foo",
                "metadata": {"name": "bar", "namespace": TEST_NAMESPACE},
            }
        ],
    )
    session = setup_session(deploy_manager=deploy_manager)
    comp = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
            ("bat", {"kind": "Baz", "metadata": {"name": "bat"}}),
        ],
    )
    comp.render_chart(session)
    assert deploy_manager.has_obj("Foo", "bar", TEST_NAMESPACE)
    assert not deploy_manager.has_obj("Baz", "bat", TEST_NAMESPACE)
    assert comp.disable(session)
    assert not deploy_manager.has_obj("Foo", "bar", TEST_NAMESPACE)
    assert not deploy_manager.has_obj("Baz", "bat", TEST_NAMESPACE)


def test_disable_before_render():
    """Make sure disable errors correctly if called before render"""
    session = setup_session()
    comp = get_comp_type()(session=session)
    with pytest.raises(AssertionError):
        comp.disable(session)


def test_disable_failure():
    """Make sure disable returns failure if the deploy_manager fails to run
    disable
    """
    session = setup_session(deploy_manager=MockDeployManager(disable_fail=True))
    comp = get_comp_type()(session=session)
    comp.render_chart(session)
    assert not comp.disable(session)


def test_disable_raise():
    """Make sure an exception thrown during disable correctly makes it out
    of the call
    """
    session = setup_session(deploy_manager=MockDeployManager(disable_raise=True))
    comp = get_comp_type()(session=session)
    comp.render_chart(session)
    with pytest.raises(AssertionError):
        comp.disable(session)


############
## verify ##
############


def test_verify_all_components_true():
    """Test that the default verify will use verify_resource to verify all
    managed objects and return True when they all verify successfully.
    """
    dm = MockDeployManager()
    with library_config(dry_run=False):
        session = setup_session(deploy_manager=dm)
        comp = get_comp_type()(
            session=session,
            api_objects=[
                ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
                (
                    "bat",
                    {"kind": "Baz", "metadata": {"name": "bat", "namespace": "diff"}},
                ),
            ],
        )
        # Render and deploy
        comp.render_chart(session)
        comp.deploy(session)

        # Make sure components are verified because they exist
        assert comp.verify(session)


def test_verify_all_components_false():
    """Test that the default verify will use verify_resource to verify all
    managed objects and return False when they all are unverified.
    """
    dm = MockDeployManager(deploy_fail=True)
    with library_config(dry_run=False):
        session = setup_session(deploy_manager=dm)
        comp = get_comp_type()(
            session=session,
            api_objects=[
                ("bar", {"kind": "Foo", "metadata": {"name": "bar"}}),
                (
                    "bat",
                    {"kind": "Baz", "metadata": {"name": "bat", "namespace": "diff"}},
                ),
            ],
        )
        # Render and deploy
        comp.render_chart(session)
        comp.deploy(session)

        # Make sure resources are not verified because they didn't get deployed
        assert not comp.verify(session)


###########
## other ##
###########


def test_to_dict():
    """Make sure that the internal resource name is formatted correctly.
    This one is also pretty basic...
    """
    comp_name = "foo"
    session = setup_session()
    component = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": comp_name}}),
        ],
    )
    comp_dict = component.to_dict(session)

    assert comp_dict[0]["metadata"]["name"] == comp_name

    # Call it again to exercise the line that avoids re-calling render
    comp_dict2 = component.to_dict(session)
    assert comp_dict == comp_dict2


def test_to_config():
    """Make sure that the internal resource name is formatted correctly.
    This one is also pretty basic...
    """
    comp_name = "foo"
    session = setup_session()
    component = get_comp_type()(
        session=session,
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": comp_name}}),
        ],
    )
    comp_config = component.to_config(session)

    assert comp_config[0].metadata.name == comp_name


def test_str():
    """Just make sure it doesn't raise!"""
    comp_name = "foo"
    component = get_comp_type(name=comp_name)(session=setup_session())
    assert comp_name in str(component)


def test_to_file():
    """Make sure that serializing to a file works cleanly.

    NOTE: At this point, file serialization is only intended to create files
        that can be subsequently applied outside of oper8, so we don't worry
        about safe round-tripping for things like dates and "on/off" style
        bools.
    """
    comp_name = "foo"
    component = get_comp_type()(
        session=setup_session(),
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": comp_name}}),
        ],
    )
    with tempfile.TemporaryDirectory() as test_dir:
        with library_config(working_dir=test_dir):
            session = setup_session()
            comp_objects = component.to_dict(session)
            comp_file = component.to_file(session)
            assert os.path.exists(comp_file)
            with open(comp_file) as f:
                comp_objects2 = list(yaml.safe_load_all(f))
            assert comp_objects2 == comp_objects


def test_to_file_disabled():
    """Make sure that serializing to a file handles disabled components
    correctly by either not rendering them or rendering them in a special dir.
    """
    comp_name = "foo"
    component = get_comp_type()(
        session=setup_session(),
        api_objects=[
            ("bar", {"kind": "Foo", "metadata": {"name": comp_name}}),
        ],
        disabled=True,
    )

    with tempfile.TemporaryDirectory() as test_dir:
        # Test with dumping disabled
        with library_config(working_dir=test_dir, dump_disabled=False):
            session = setup_session()
            assert component.to_file(session) is None

        # Test with dumping enabled
        with library_config(working_dir=test_dir, dump_disabled=True):
            session = setup_session()
            comp_file = component.to_file(session)
            assert "DISABLED" in comp_file
            assert os.path.exists(comp_file)


def test_managed_object_repr():
    """Just make sure it doesn't raise!"""
    obj = ManagedObject(
        {
            "apiVersion": "v1",
            "kind": "Foo",
            "metadata": {"name": "foo", "namespace": "bar"},
        }
    )
    repr(obj)


def test_managed_object_eq():
    """Test that equality ignores definition"""
    obj_a = ManagedObject(
        {
            "apiVersion": "v1",
            "kind": "Foo",
            "metadata": {"name": "foo", "namespace": "bar"},
        }
    )
    obj_b = ManagedObject(
        {
            "apiVersion": "v1",
            "kind": "Foo",
            "metadata": {"name": "foo", "namespace": "bar"},
        }
    )
    obj_c = ManagedObject(
        {
            "apiVersion": "v1",
            "kind": "Foo",
            "metadata": {"name": "bar", "namespace": "bar"},
        }
    )
    obj_d = ManagedObject(
        {
            "apiVersion": "v1",
            "kind": "Baz",
            "metadata": {"name": "foo", "namespace": "bar"},
        }
    )
    obj_e = ManagedObject(
        {
            "apiVersion": "v2",
            "kind": "Foo",
            "metadata": {"name": "foo", "namespace": "bar"},
        }
    )
    assert obj_a == obj_b
    assert obj_a != obj_c
    assert obj_a != obj_d
    assert obj_a != obj_e



"""Import the ThreadBase and subclasses"""
# Local
from .base import ThreadBase
from .heartbeat import HeartbeatThread
from .reconcile import ReconcileThread
from .timer import TimerThread
from .watch import WatchThread



"""
This module holds shared functionality for adding dependency annotations to all
resources that need them.

A dependency annotation on a Pod encodes a unique hash of the set of
data-resources that the Pod depends on. For example, if a Pod mounds a Secret
and a ConfigMap, the dependency annotation will hold a unique hash of the data
content of these secrets. The role of the dependency annotation is to force a
rollover when upstream data-resources change their content so that the content
is guaranteed to be picked up by the consuming Pod.
"""

# Standard
from typing import List, Optional, Tuple, Union
import hashlib
import json

# First Party
import alog

# Local
from .constants import DEPS_ANNOTATION
from oper8 import Component, Session
from oper8.session import _SESSION_NAMESPACE
from oper8.utils import merge_configs

log = alog.use_channel("DEPS")

## Common Functions ############################################################


@alog.logged_function(log.debug)
def add_deps_annotation(
    component: Component,
    session: Session,
    resource_definition: dict,
) -> dict:
    """Add the dependency hash annotation to any pods found in the given object

    Args:
        component:  Component
            The component that this resource belongs to
        session:  Session
            The session for this deploy
        resource_definition:  dict
            The dict representation of the resource to modify

    Returns:
        resource_definition:  dict
            The dict representation of the resource with any modifications
            applied
    """
    resource_name = "{}/{}".format(
        resource_definition.get("kind"),
        resource_definition.get("metadata", {}).get("name"),
    )

    # Look for any/all pod annotations
    pod = _find_pod(resource_definition)
    if pod is not None:
        log.debug2("Found Pod for [%s]", resource_name)
        log.debug4(pod)

        # Traverse through and look for anything that looks like a secret or
        # configmap reference
        deps_map = _find_pod_data_deps(pod)
        log.debug3("Deps Map: %s", deps_map)
        if deps_map:
            # Go through each dependency and determine if it needs to be fetched
            # of if it's part of the owning component
            deps_list = []
            for dep_kind, dep_names in deps_map.items():
                for dep_name in dep_names:
                    # Look for this object in the objects managed by this
                    # component.
                    #
                    # NOTE: This will only be the components which have been
                    #   declared earlier in the chart or have explicitly been
                    #   marked as upstreams of this object.
                    found_in_component = False
                    for obj in component.managed_objects:
                        log.debug4("Checking %s/%s", obj.kind, obj.name)
                        if obj.kind == dep_kind and obj.name == dep_name:
                            log.debug3(
                                "Found intra-chart dependency of %s: %s",
                                resource_name,
                                obj,
                            )
                            deps_list.append(obj.definition)
                            found_in_component = True
                            break

                    # If not found in the component, add it as a lookup
                    if not found_in_component:
                        log.debug3(
                            "Found extra-chart dependency of %s: %s/%s",
                            resource_name,
                            dep_kind,
                            dep_name,
                        )
                        deps_list.append((dep_kind, dep_name))

            # Add the annotation with the full list
            md = pod.setdefault("metadata", {})
            annos = md.setdefault("annotations", {})
            md["annotations"] = merge_configs(
                annos, get_deps_annotation(session, deps_list, resource_name)
            )

    log.debug4("Updated Definition of [%s]: %s", resource_name, resource_definition)
    return resource_definition


def get_deps_annotation(
    session: Session,
    dependencies: List[Union[dict, Tuple[str, str]]],
    resource_name: str = "",
    namespace: Optional[str] = _SESSION_NAMESPACE,
) -> dict:
    """Get a dict holding an annotation key/value pair representing the unique
    content hash of all given dependencies. This can be used to force pods to
    roll over when a dependency such as a ConfigMap or Secret changes its
    content. This function supports two ways of fetching dependency content:

    1. Dict representation of the object
    2. Tuple of the scoped (kind, name) for the object

    Additionally, this function holds special logic for ConfigMap and Secret
    dependencies, but can handle arbitrary kinds. For kinds without special
    logic, the full dict representation is used to compute the hash.

    Args:
        session:  Session
            The current session
        dependencies:  list(dict or str or cdk8s.ApiObject)
            An ordered list of dependencies to compute the content hash from
        resource_name:  str
            A string name for the resource (used for logging)
        namespace:  Optional[str]
            Namespace where the dependencies live. Defaults to session.namespace

    Returns:
        deps_annotation:  dict
            A dict representation of the key/value pair used to hold the content
            hash for the given set of dependencies
    """
    content_hash = hashlib.sha1()
    namespace = namespace if namespace != _SESSION_NAMESPACE else session.namespace
    for dep in dependencies:
        # Get the dict representation depending on what type this is
        if isinstance(dep, tuple):
            log.debug3("[%s] Handling tuple dependency: %s", resource_name, dep)
            assert len(dep) == 2, f"Invalid dependency tuple given: {dep}"
            kind, name = dep
            success, dep_dict = session.get_object_current_state(
                name=name,
                kind=kind,
                namespace=namespace,
            )
            assert success, f"Failed to fetch current state of {kind}/{name}"

            # There are several reasons that the upstream dependency would not
            # be found, some legitimate and some not:
            #
            # 1. The dependency is not managed by this operator and this is a
            #   dry run. This can't be solved since we don't have control over
            #   the state of the cluster in dry run.
            #
            # 2. The dependency is part of a cyclic dependency between
            #   Components. While a sign of something bad, this is ultimately
            #   something that needs to be solved by decoupling the Component
            #   dependencies.
            #
            # 3. The upstream is an undeclared chart dependency. This is an
            #   easily fixed bug in the component by adding the necessary
            #   add_dependency() calls.
            #
            # 4. The upstream is part of an undeclared component dependency.
            #   This is an easily fixed bug in the parent Application by adding
            #   the missing add_component_dependency() calls.
            #
            # Since some of these are things that should be quickly fixed, but
            # some are signs of larger systemic problems, we warn and move on.
            # For (1), these external dependencies should be present in the
            # cluster. For the rest, once the deploy completes for the coupled
            # components, the resources will show up and the next reconcile will
            # cause the hash to change to what it should be.
            if dep_dict is None:
                log.warning(
                    "Working around missing external data dependency for [%s]: %s/%s",
                    resource_name,
                    kind,
                    name,
                )
                continue
        else:
            log.debug3(
                "[%s] Handling dict dependency: %s",
                resource_name,
                dep.get("metadata", {}).get("name"),
            )
            assert isinstance(dep, dict), f"Unknown dependency type: {type(dep)}"
            dep_dict = dep

        # The hash should be unique to the name and kind
        kind = dep_dict.get("kind", "")
        name = dep_dict.get("metadata", {}).get("name", "")
        content_hash.update(kind.encode("utf-8"))
        content_hash.update(name.encode("utf-8"))

        # Compute the data hash based on any kind-specific logic
        if kind in ["Secret", "ConfigMap"]:
            log.debug2("Getting data hash for dep of kind %s", kind)
            data_dict = dep_dict.get("data", {})
        else:
            log.debug2("Getting full hash for dep of kind %s", kind)
            data_dict = dep_dict
        log.debug4("Data Dict: %s", data_dict)

        # Add to the overall hash
        content_hash.update(json.dumps(data_dict, sort_keys=True).encode("utf-8"))

    # Return the annotation dict
    final_hash = content_hash.hexdigest()
    log.debug2("[%s] Final Hash: %s", resource_name, final_hash)
    return {DEPS_ANNOTATION: final_hash}


## Implementation Details ######################################################


def _find_pod(resource_definition: dict) -> dict:
    """Look through the object and return a refernce to any pod resource or
    template found
    """
    kind = resource_definition.get("kind")
    log.debug2("Looking for pod annotations for %s", kind)

    # Pod
    if kind == "Pod":
        return resource_definition

    # Deployment, ReplicaSet, StatefulSet
    if kind in ["Deployment", "ReplicaSet", "StatefulSet"]:
        return resource_definition.setdefault("spec", {}).setdefault("template", {})

    # No pod annotations found
    log.debug("No pod annotations found for [%s]", kind)
    return None


def _find_pod_data_deps(pod: dict) -> dict:
    """Look through a pod's definition for any references to Secret or ConfigMap
    resources
    """
    log.debug4("Looking for deps in: %s", pod)
    pod_spec = pod.get("spec", {})

    deps_map = {}

    # volumes
    for volume in pod_spec.get("volumes", []):
        # Secret volume
        secret_name = volume.get("secret", {}).get("secretName")
        if secret_name:
            log.debug2("Found Secret volume dependency: %s", secret_name)
            deps_map.setdefault("Secret", set()).add(secret_name)

        # ConfigMap volume
        cm_name = volume.get("configMap", {}).get("name")
        if cm_name:
            log.debug2("Found ConfigMap volume dependency: %s", cm_name)
            deps_map.setdefault("ConfigMap", set()).add(cm_name)

    # env
    for container in pod_spec.get("containers", []):
        for env_var in container.get("env", []):
            value_from = env_var.get("valueFrom", {})
            if value_from:
                # Secret reference
                secret_name = value_from.get("secretKeyRef", {}).get("name")
                if secret_name:
                    log.debug2("Found Secret env dependency: %s", secret_name)
                    deps_map.setdefault("Secret", set()).add(secret_name)

                # ConfigMap reference
                cm_name = value_from.get("configMapKeyRef", {}).get("name")
                if cm_name:
                    log.debug2("Found ConfigMap env dependency: %s", cm_name)
                    deps_map.setdefault("ConfigMap", set()).add(cm_name)

    # Return the set of named deps types
    return {dep_type: sorted(list(deps)) for dep_type, deps in deps_map.items()}



"""
The DatastoreSingletonFactoryBase class defines the common functionality that
all datastore type factories will use. It implements common logic for
constructing named singleton instances of a given datastore type.
"""

# Standard
from typing import Optional

# First Party
import alog

# Local
from ... import Component, Session
from ...utils import merge_configs
from ..utils import constants
from .connection_base import DatastoreConnectionBase
from .interfaces import Datastore

## Factory Base ################################################################

log = alog.use_channel("FCTRY")


class classproperty:
    """@classmethod+@property
    CITE: https://stackoverflow.com/a/22729414
    """

    def __init__(self, func):
        self.func = classmethod(func)

    def __get__(self, *args):
        return self.func.__get__(*args)()


class DatastoreSingletonFactoryBase:
    """The DatastoreSingletonFactoryBase manages instances of all datastore
    types as singletons on a per-deployment basis. It provides functionality for
    derived classes to define a specific DATASTORE_TYPE (e.g. redis) and
    register implementations of that type.

    The instances of each type are held as singletons scoped to the individual
    deployment (session.deploy_id). This is done to support multiple calls to
    fetch a named instance within a given deployment without reconstructing, but
    to allow configuration to change between deploys.
    """

    ## Private Members #########################################################

    # Singleton dict of constructors for each implementation type
    _type_constructors = {}

    # Singleton dict of named components
    _components = {}

    # Singleton dict of named connections
    _connections = {}

    # Class attribute that all individual factory types must have.
    # NOTE: This will be used as the key in the CR's datastores section
    _DATASTORE_TYPE_ATTRIBUTE_NAME = "DATASTORE_TYPE"

    # Class attribute that must be defined on an implementation to define the
    # common connection type
    _CONNECTION_TYPE_ATTRIBUTE = "CONNECTION_TYPE"

    ## Public interface ########################################################

    @classproperty
    def datastore_type(cls):
        return getattr(cls, cls._DATASTORE_TYPE_ATTRIBUTE_NAME)

    @classproperty
    def connection_type(cls):
        return getattr(cls, cls._CONNECTION_TYPE_ATTRIBUTE)

    @classmethod
    def get_component(
        cls,
        session: Session,
        name: Optional[str] = None,
        disabled: bool = False,
        config_overrides: Optional[dict] = None,
    ) -> Optional[Component]:
        """Construct an instance of the datastore type's component

        Args:
            session:  Session
                The session for the current deployment
            name:  Optional[str]
                The name of the singleton instance to get. If not provided, a
                top-level instance is used (e.g. datastores.postgres.type)
            disabled:  bool
                Whether or not the component is disabled in this deployment
            config_overrides:  Optional[dict]
                Optional runtime config values. These will overwrite any values
                pulled from the session.config

        Returns:
            instance:  Optional[Component]
                The constructed component if one is needed
        """
        return cls._get_component(
            session=session,
            name=name,
            disabled=disabled,
            config_overrides=config_overrides,
        )

    @classmethod
    def get_connection(
        cls,
        session: Session,
        name: Optional[str] = None,
        allow_from_component: bool = True,
    ) -> DatastoreConnectionBase:
        """Get the connection details for a named instance of the datastore type

        If not pre-constructed by the creation of the Component, connection
        details are pulled from the CR directly
        (spec.datastores.<datastore_type>.[<name>].connection)

        Args:
            session:  Session
                The session for the current deployment
            name:  Optional[str]
                The name of the singleton instance to get. If not provided, a
                top-level instance is used
            allow_from_component:  bool
                If True, use connection info from the component

        Returns:
            connection:  DatastoreConnectionBase
                The connection for this instance
        """
        return cls._get_connection(session, name, allow_from_component)

    @classmethod
    def register_type(cls, type_class: Datastore):
        """Register a new type constructor

        Args:
            type_class:  Datastore
                The class that will be constructed with the config for
        """
        cls._validate_class_attributes()

        assert issubclass(
            type_class, Datastore
        ), "Datastore types use component_class=Datastore"
        datastore_type_classes = cls._type_constructors.setdefault(
            cls.datastore_type, {}
        )
        if type_class.TYPE_LABEL in datastore_type_classes:
            log.warning("Got duplicate registration for %s", type_class.TYPE_LABEL)
        datastore_type_classes[type_class.TYPE_LABEL] = type_class

    ## Implementation Details ##################################################

    @classmethod
    def _validate_class_attributes(cls):
        """Since this class is always used statically, this helper makes sure
        the expected class attributes
        """
        assert isinstance(
            getattr(
                cls, DatastoreSingletonFactoryBase._DATASTORE_TYPE_ATTRIBUTE_NAME, None
            ),
            str,
        ), "Incorrectly configured datastore [{}]. Must define str [{}]".format(
            cls, DatastoreSingletonFactoryBase._DATASTORE_TYPE_ATTRIBUTE_NAME
        )
        connection_type = getattr(
            cls, DatastoreSingletonFactoryBase._CONNECTION_TYPE_ATTRIBUTE, None
        )
        assert isinstance(connection_type, type) and issubclass(
            connection_type, DatastoreConnectionBase
        ), (
            f"Incorrectly configured datastore [{cls}]. Must define "
            f"[{DatastoreSingletonFactoryBase._CONNECTION_TYPE_ATTRIBUTE}] as "
            "a DatastoreConnectionBase type"
        )

    @classmethod
    def _get_component(
        cls,
        session: Session,
        name: Optional[str] = None,
        disabled: bool = False,
        config_overrides: Optional[dict] = None,
        allow_instantiation: bool = True,
    ) -> Optional[Component]:
        """Implementation detail for get_component which can be called by
        get_connection and disallow lazy creation of the singleton.
        """
        cls._validate_class_attributes()

        # First, check to see if there's a connection already available based on
        # conneciton details in the CR. If so, we won't create the component
        conn = cls._get_connection(session, name, allow_from_component=False)
        if conn is not None:
            log.debug(
                "Found connection for [%s] in the CR. Not constructing the component.",
                cls.datastore_type,
            )
            return None

        # Get the pre-existing instances for this datastore type (keyed by the
        # datastore subclass)
        datastore_components = cls._components.setdefault(cls.datastore_type, {})

        # Get the app config section for this instance by name
        instance_config = merge_configs(
            session.config.get(cls.datastore_type, {}),
            config_overrides or {},
        )
        log.debug4("Full config: %s", instance_config)
        if name is not None:
            instance_config = instance_config.get(name)
        assert (
            instance_config is not None
        ), f"Cannot construct unknown [{cls.datastore_type}] instance: {name}"
        assert (
            "type" in instance_config
        ), f"Missing required [type] key for [{cls.datastore_type}/{name}]"

        # Fetch the current instance/deploy_id
        instance, deploy_id = datastore_components.get(name, (None, None))

        # If the deploy_id has changed, remove any current instance
        if deploy_id != session.id:
            instance = None
            datastore_components.pop(name, None)

        # If there is not a valid instance and it's allowed, construct it
        if not instance and allow_instantiation:
            log.debug2(
                "Constructing [%s]/%s for the first time for deploy [%s]",
                cls.datastore_type,
                name,
                session.id,
            )
            type_key = instance_config.type

            # Fetch the class for this type of the datastore
            datastore_type_classes = cls._type_constructors.get(cls.datastore_type, {})
            type_class = datastore_type_classes.get(type_key)
            assert (
                type_class is not None
            ), f"Cannot construct unsupported type [{type_key}]"

            # If there is a name provided, create a wrapper component with the
            # given name
            if name is not None:
                instance_class_name = f"{type_class.name}-{name}"
                log.debug2("Wrapping %s with instance name override", type_class)

                class InstanceClass(type_class):
                    """Wrapper for {}/{} with instance naming""".format(
                        cls.datastore_type, type_key
                    )

                    name = instance_class_name

            else:
                log.debug2("No instance name wrapping needed for %s", type_class)
                InstanceClass = type_class
            log.debug("Constructing %s", type_key)
            instance = InstanceClass(
                session=session,
                config=instance_config,
                instance_name=name,
                disabled=disabled,
            )
            datastore_components[name] = (instance, session.id)

        # Return the singleton
        return instance

    @classmethod
    def _get_connection(
        cls,
        session: Session,
        name: Optional[str] = None,
        allow_from_component: bool = True,
    ) -> Optional[DatastoreConnectionBase]:
        """Implementation for get_connection that can be used by _get_component
        to fetch connections from the CR
        """
        cls._validate_class_attributes()

        # Get the pre-existing instances for this datastore type (keyed by the
        # datastore subclass)
        connection, deploy_id = cls._connections.get(cls.datastore_type, {}).get(
            name, (None, None)
        )

        # If there is no connection for this deploy already, deserialize it from
        # the CR
        if connection is None or deploy_id != session.id:
            log.debug("Constructing %s connection from config", cls.datastore_type)

            # Get the CR config for this datastore type
            ds_config = session.spec.get(constants.SPEC_DATASTORES, {}).get(
                cls.datastore_type, {}
            )
            if name is not None:
                ds_config = ds_config.get(name, {})
            ds_config = ds_config.get(constants.SPEC_DATASTORE_CONNECTION)
            log.debug3(
                "%s/%s connection config: %s", cls.datastore_type, name, ds_config
            )

            if ds_config is not None:
                # Deserialize connection from sub-cr connection specification
                connection = cls.connection_type.from_dict(session, ds_config)
                cls._connections.setdefault(cls.datastore_type, {})[name] = (
                    connection,
                    session.id,
                )
            elif allow_from_component:
                # Add the connection information for this instance
                instance = cls._get_component(session, name, allow_instantiation=False)
                assert (
                    instance is not None
                ), f"No instance or config available for {cls.datastore_type}"
                connection = instance.get_connection()
                assert isinstance(
                    connection, cls.connection_type
                ), f"Got incorrect [{cls.datastore_type}] connection type: {type(connection)}"
                cls._connections.setdefault(cls.datastore_type, {})[name] = (
                    connection,
                    session.id,
                )
            else:
                log.debug2(
                    "No connection details for %s found in CR", cls.datastore_type
                )
                return None

        # Return the connection singleton
        return cls._connections[cls.datastore_type][name][0]



"""
The TimerThread is a helper class used to run schedule events
"""

# Standard
from datetime import datetime
from heapq import heappop, heappush
from typing import Any, Callable, Dict, List, Optional
import queue
import threading

# First Party
import alog

# Local
from ..utils import MIN_SLEEP_TIME, Singleton, TimerEvent
from .base import ThreadBase

log = alog.use_channel("TMRTHRD")


class TimerThread(ThreadBase, metaclass=Singleton):
    """The TimerThread class is a helper class to run scheduled actions. This is very similar
    to threading.Timer stdlib class except that it uses one shared thread for all events
    instead of a thread per event."""

    def __init__(self, name: Optional[str] = None):
        """Initialize a priorityqueue like object and a synchronization object"""
        super().__init__(name=name or "timer_thread", daemon=True)

        # Use a heap queue instead of a queue.PriorityQueue as we're already handling
        # synchronization with the notify condition
        # https://docs.python.org/3/library/heapq.html?highlight=heap#priority-queue-implementation-notes
        self.timer_heap = []
        self.notify_condition = threading.Condition()

    def run(self):
        """The TimerThread's control loop sleeps until the next schedule
        event and executes all pending actions."""
        if not self.check_preconditions():
            return

        while True:
            # Wait until the next event or a new event is pushed
            with self.notify_condition:
                time_to_sleep = self._get_time_to_sleep()
                if time_to_sleep:
                    log.debug2(
                        "Timer waiting %ss until next scheduled event", time_to_sleep
                    )
                else:
                    log.debug2("Timer waiting until event queued")
                self.notify_condition.wait(timeout=time_to_sleep)

            if not self.check_preconditions():
                return

            # Get all the events to be executed
            event_list = self._get_all_current_events()
            for event in event_list:
                log.debug("Timer executing action for event: %s", event)
                event.action(*event.args, **event.kwargs)

    ## Class Interface ###################################################

    def stop_thread(self):
        """Override stop_thread to wake the control loop"""
        super().stop_thread()
        # Notify timer thread of shutdown
        log.debug2("Acquiring notify condition for shutdown")
        with self.notify_condition:
            log.debug("Notifying TimerThread of shutdown")
            self.notify_condition.notify_all()

    ## Public Interface ###################################################

    def put_event(
        self, time: datetime, action: Callable, *args: Any, **kwargs: Dict
    ) -> Optional[TimerEvent]:
        """Push an event to the timer

        Args:
            time: datetime
                The datetime to execute the event at
            action: Callable
                The action to execute
            *args: Any
                Args to pass to the action
            **kwargs: Dict
                Kwargs to pass to the action

        Returns:
            event: Optional[TimerEvent]
                TimerEvent describing the event and can be cancelled
        """
        # Don't allow pushing to a stopped thread
        if self.should_stop():
            return None

        # Create a timer event and push it to the heap
        event = TimerEvent(time=time, action=action, args=args, kwargs=kwargs)
        with self.notify_condition:
            heappush(self.timer_heap, event)
            self.notify_condition.notify_all()
        return event

    ## Time Functions  ###################################################

    def _get_time_to_sleep(self) -> Optional[int]:
        """Calculate the time to sleep based on the current queue

        Returns:
            time_to_wait: Optional[int]
               The time to wait if there's an object in the queue"""
        with self.notify_condition:
            obj = self._peak_next_event()
            if obj:
                time_to_sleep = (obj.time - datetime.now()).total_seconds()
                if time_to_sleep < MIN_SLEEP_TIME:
                    return MIN_SLEEP_TIME
                return time_to_sleep

            return None

    ## Queue Functions  ###################################################

    def _get_all_current_events(self) -> List[TimerEvent]:
        """Get all the current events that should execute

        Returns:
            current_events: List[TimerEvent]
                List of timer events to execute
        """
        event_list = []
        # With lock preview the next object
        with self.notify_condition:
            while len(self.timer_heap) != 0:
                obj_preview = self._peak_next_event()
                # If object exists and should've already executed then remove object from queue
                # and add it to return list
                if obj_preview and obj_preview.time < datetime.now():
                    try:
                        obj = heappop(self.timer_heap)
                        if obj.stale:
                            log.debug2("Skipping timer event %s", obj)
                            continue
                        event_list.append(obj)
                    except queue.Empty:
                        break
                else:
                    break
        return event_list

    def _peak_next_event(self) -> Optional[TimerEvent]:
        """Get the next timer event without removing it from the queue

        Returns:
            next_event: TimerEvent
                The next timer event if one exists
        """
        with self.notify_condition:
            if self.timer_heap:
                return self.timer_heap[0]
            return None



""" __init__ file for Filter submodule. Imports all filters, functions,
 and classes from filters module """
# Local
from .common import get_configured_filter, get_filters_for_resource_id
from .filters import (
    AnnotationFilter,
    CreationDeletionFilter,
    DependentWatchFilter,
    DisableFilter,
    EnableFilter,
    Filter,
    GenerationFilter,
    LabelFilter,
    NoGenerationFilter,
    PauseFilter,
    ResourceVersionFilter,
    SubsystemStatusFilter,
    UserAnnotationFilter,
)
from .manager import AndFilter, FilterManager, OrFilter



"""
CompletionState holds info about how a DAG Runner completes
"""


# Standard
from typing import List, Optional

# First Party
import alog

# Local
from .node import Node

log = alog.use_channel("DAG")

## Completion state ##############################################################


class CompletionState:
    """
    This class holds the definition of a CompletionState which manages all
    the information about how the nodes in a rollout Runner terminated
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        verified_nodes: Optional[List[Node]] = None,
        unverified_nodes: Optional[List[Node]] = None,
        failed_nodes: Optional[List[Node]] = None,
        unstarted_nodes: Optional[List[Node]] = None,
        exception: Optional[Exception] = None,
    ):
        """Construct with each node set"""
        self.verified_nodes = set(verified_nodes or [])
        self.unverified_nodes = set(unverified_nodes or [])
        self.failed_nodes = set(failed_nodes or [])
        self.unstarted_nodes = set(unstarted_nodes or [])
        self.all_nodes = (
            self.verified_nodes.union(self.unverified_nodes)
            .union(self.failed_nodes)
            .union(self.unstarted_nodes)
        )
        self.exception = exception

        # Make sure the sets are not overlapping
        sets = [
            self.verified_nodes,
            self.unverified_nodes,
            self.failed_nodes,
            self.unstarted_nodes,
        ]
        for i, node_set_a in enumerate(sets):
            for j, node_set_b in enumerate(sets):
                if i != j:
                    assert not node_set_a.intersection(node_set_b), (
                        "Programming Error: "
                        + f"CompletionState constructed with overlapping sets: {str(self)}"
                    )

    def __str__(self):
        return "\n".join(
            [
                f"[NODES] {key}: {list(sorted(nodes))}"
                for key, nodes in [
                    ("Verified", [node.get_name() for node in self.verified_nodes]),
                    ("Unverified", [node.get_name() for node in self.unverified_nodes]),
                    ("Failed", [node.get_name() for node in self.failed_nodes]),
                    ("Unstarted", [node.get_name() for node in self.unstarted_nodes]),
                ]
            ]
            + [
                f"Exception: {self.exception}",
            ]
        )

    def __eq__(self, other: "CompletionState"):
        return (
            self.verified_nodes == other.verified_nodes
            and self.unverified_nodes == other.unverified_nodes
            and self.failed_nodes == other.failed_nodes
            and self.unstarted_nodes == other.unstarted_nodes
        )

    def deploy_completed(self) -> bool:
        """Determine if the dag completed all nodes through to the deploy
        step

        NOTE: An empty node set is considered completed

        Returns:
            completed:  bool
                True if there are no failed nodes and no unstarted nodes
        """
        return not self.failed_nodes and not self.unstarted_nodes

    def verify_completed(self) -> bool:
        """Determine if the dag completed all nodes through to the verification
        step

        NOTE: An empty node set is considered verified

        Returns:
            completed:  bool
                True if there are no nodes found outside of the verified_nodes
                and there is no exception in the termination state
        """
        return (
            not self.unverified_nodes
            and not self.failed_nodes
            and not self.unstarted_nodes
            and not self.exception
        )

    def failed(self) -> bool:
        """Determine if any of the nodes failed

        Returns:
            failed:  bool
                True if there are any nodes in the failed state or there is a
                fatal error
        """
        return bool(self.failed_nodes) or self._fatal_exception()

    def _fatal_exception(self):
        """Helper to determine if there is a fatal exception in the state"""
        return self.exception is not None and getattr(
            self.exception, "is_fatal_error", True
        )



"""
Common utilities shared across components in the library
"""

# Standard
from typing import Any
import copy
import datetime
import inspect

# Third Party
import six

# First Party
import aconfig
import alog

# Local
from . import config, constants
from .dag import ResourceNode
from .exceptions import assert_cluster

log = alog.use_channel("OPUTL")


# Forward declaration for Session
SESSION_TYPE = "Session"

# Sentinel for missing dict values
__MISSING__ = "__MISSING__"

## Dicts #######################################################################


def merge_configs(base, overrides) -> dict:
    """Helper to perform a deep merge of the overrides into the base. The merge
    is done in place, but the resulting dict is also returned for convenience.

    The merge logic is quite simple: If both the base and overrides have a key
    and the type of the key for both is a dict, recursively merge, otherwise
    set the base value to the override value.

    Args:
        base:  dict
            The base config that will be updated with the overrides
        overrides:  dict
            The override config

    Returns:
        merged:  dict
            The merged results of overrides merged onto base
    """
    for key, value in overrides.items():
        if (
            key not in base
            or not isinstance(base[key], dict)
            or not isinstance(value, dict)
        ):
            base[key] = value
        else:
            base[key] = merge_configs(base[key], value)

    return base


def nested_set(dct: dict, key: str, val: Any):
    """Helper to set values in a dict using 'foo.bar' key notation

    Args:
        dct:  dict
            The dict into which the key will be set
        key:  str
            Key that may contain '.' notation indicating dict nesting
        val:  Any
            The value to place at the nested key
    """
    parts = key.split(constants.NESTED_DICT_DELIM)
    for i, part in enumerate(parts[:-1]):
        dct = dct.setdefault(part, {})
        if not isinstance(dct, dict):
            raise TypeError(
                "Intermediate key {} is not a dict".format(  # pylint: disable=consider-using-f-string
                    constants.NESTED_DICT_DELIM.join(parts[:i])
                )
            )
    dct[parts[-1]] = val


def nested_get(dct: dict, key: str, dflt=None) -> Any:
    """Helper to get values from a dict using 'foo.bar' key notation

    Args:
        dct:  dict
            The dict into which the key will be set
        key:  str
            Key that may contain '.' notation indicating dict nesting

    Returns:
        val:  Any
            Whatever is found at the given key or None if the key is not found.
            This includes missing intermediate dicts.
    """
    parts = key.split(constants.NESTED_DICT_DELIM)
    for i, part in enumerate(parts[:-1]):
        dct = dct.get(part, __MISSING__)
        if dct is __MISSING__:
            return dflt
        if not isinstance(dct, dict):
            raise TypeError(
                "Intermediate key {} is not a dict".format(  # pylint: disable=consider-using-f-string
                    constants.NESTED_DICT_DELIM.join(parts[:i])
                )
            )
    return dct.get(parts[-1], dflt)


# Stolen from kubernetes but modified for None pruning and safety
# https://github.com/kubernetes-client/python/blob/d67bc8c2bdb89b29c17c1ba0edb03a48d977c0e2/kubernetes/client/api_client.py#L202
def sanitize_for_serialization(obj):  # pylint: disable=too-many-return-statements
    """Builds a JSON POST object.
    If obj is None, return None.
    If obj is str, int, long, float, bool, return directly.
    If obj is datetime.datetime, datetime.date
        convert to string in iso8601 format.
    If obj is list, sanitize each element in the list.
    If obj is dict, return the dict.
    If obj is OpenAPI model, return the properties dict.
    :param obj: The data to serialize.
    :return: The serialized form of data.
    """
    if obj is None:  # pylint: disable=no-else-return
        return None
    elif isinstance(obj, (float, bool, bytes, six.text_type) + six.integer_types):
        return obj
    elif isinstance(obj, list):
        return [sanitize_for_serialization(sub_obj) for sub_obj in obj]
    elif isinstance(obj, tuple):
        return tuple(sanitize_for_serialization(sub_obj) for sub_obj in obj)
    elif isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    elif isinstance(obj, ResourceNode):
        return sanitize_for_serialization(obj.manifest)
    elif isinstance(obj, property):
        return sanitize_for_serialization(obj.fget())

    if isinstance(obj, dict):
        obj_dict = obj
    elif hasattr(obj, "attribute_map"):
        # Convert model obj to dict except
        # `openapi_types` and `attribute_map`.
        # Convert attribute name to json key in
        # model definition for request.
        obj_dict = {}
        for attr, name in six.iteritems(obj.attribute_map):
            if hasattr(obj, attr):
                obj_dict[name] = getattr(obj, attr)

    # Prune fields which are None but keep
    # empty arrays or dictionaries
    return_dict = {}
    for key, val in six.iteritems(obj_dict):
        updated_obj = sanitize_for_serialization(val)
        if updated_obj is not None:
            return_dict[key] = updated_obj
    return return_dict


## Common Usage Patterns #######################################################


def get_passthrough_annotations(session):
    """This helper gets the set of annotations that should be passed from a
    parent CR to a child subsystem CR.

    Args:
        session:  DeploySession
            The session for the current deploy

    Returns:
        annotations:  Dict[str, str]
            The dict mapping of annotations that should be passed through
    """
    annotations = session.metadata.get("annotations", {})
    passthrough_annotations = {
        k: v for k, v in annotations.items() if k in constants.ALL_ANNOTATIONS
    }

    log.debug2("Oper8 passthrough annotations: %s", passthrough_annotations)
    return passthrough_annotations


def get_manifest_version(cr_manifest: aconfig.Config) -> str:
    """Get the version for a given custom resource or from the config
    if version override provided

    Args:
        cr_manifest: aconfig.Config
            The manifest to pull the version from

    Returns:
        version: str
            The current version
    """
    if config.vcs.version_override:
        return config.vcs.version_override
    return nested_get(cr_manifest, config.vcs.field)


def add_finalizer(session: SESSION_TYPE, finalizer: str):
    """This helper adds a finalizer to current session CR

    Args:
        session:  Session
            The session for the current deploy
        finalizer: str
            The finalizer to be added
    """
    if finalizer in session.finalizers:
        return

    log.debug("Adding finalizer: %s", finalizer)

    manifest = {
        "kind": session.kind,
        "apiVersion": session.api_version,
        "metadata": copy.deepcopy(session.metadata),
    }
    manifest["metadata"].setdefault("finalizers", []).append(finalizer)
    success, _ = session.deploy_manager.deploy([manifest])

    # Once successfully added to cluster than add it to session
    assert_cluster(success, f"Failed add finalizer {finalizer}")
    session.finalizers.append(finalizer)


def remove_finalizer(session: SESSION_TYPE, finalizer: str):
    """This helper gets removes a finalizer from the current session controller

    Args:
        session:  Session
            The session for the current deploy
        finalizer: str
            The finalizer to remove

    Returns:
        annotations:  Dict[str, str]
            The dict mapping of annotations that should be passed through
    """
    if finalizer not in session.finalizers:
        return

    log.debug("Removing finalizer: %s", finalizer)

    # Create manifest with only required fields
    manifest = {
        "kind": session.kind,
        "apiVersion": session.api_version,
        "metadata": copy.deepcopy(session.metadata),
    }

    # Check to see if the object exists in the cluster
    success, found = session.get_object_current_state(
        kind=session.kind,
        api_version=session.api_version,
        name=session.name,
    )
    assert_cluster(success, "Failed to look up CR for self")

    # If still present in the cluster, update it without the finalizer
    if found:
        manifest["metadata"]["finalizers"].remove(finalizer)
        success, _ = session.deploy_manager.deploy([manifest])

        # Once successfully removed from cluster than remove from session
        assert_cluster(success, f"Failed remove finalizer {finalizer}")

    # If the finalizer has been confirmed to not be there, remove it from the
    # in-memory finalizers
    session.finalizers.remove(finalizer)


## General #####################################################################


class classproperty:  # pylint: disable=invalid-name,too-few-public-methods
    """@classmethod+@property
    CITE: https://stackoverflow.com/a/22729414
    """

    def __init__(self, func):
        self.func = classmethod(func)

    def __get__(self, *args):
        return self.func.__get__(*args)()


class abstractclassproperty:  # pylint: disable=invalid-name,too-few-public-methods
    """This decorator implements a classproperty that will raise when accessed"""

    def __init__(self, func):
        self.prop_name = func.__name__

    def __get__(self, *args):
        # If this is being called by __setattr__, we're ok because it's
        # attempting to set the attribute on the class
        curframe = inspect.currentframe()
        callframe = inspect.getouterframes(curframe, 2)[1]
        caller_name = callframe[3]
        if caller_name == "__setattr__":
            return None

        # If this is a help() call or a pdoc documentation request, return an
        # object with a docstring indicating that the property is abstract
        if (
            "help" in callframe.frame.f_code.co_names
            or callframe.frame.f_globals["__name__"] == "pdoc"
        ):

            class AbstractClassProperty:  # pylint: disable=missing-class-docstring
                __slots__ = []
                __doc__ = f"""The <{self.prop_name}> property is an abstract class property
                that must be overwritten in derived children
                """

            return AbstractClassProperty

        raise NotImplementedError(
            f"Cannot access abstractclassproperty {self.prop_name}"
        )



"""
Postgres instance factory
"""

# Local
from ..factory_base import DatastoreSingletonFactoryBase
from .connection import PostgresConnection


class PostgresFactory(DatastoreSingletonFactoryBase):
    """The common factory that will manage instances of Postgres for each deploy"""

    DATASTORE_TYPE = "postgres"
    CONNECTION_TYPE = PostgresConnection



"""
This library holds common verification routines for individual kubernetes
resources.
"""

# Standard
from datetime import datetime
from functools import partial
from typing import Callable, List, Optional

# Third Party
import dateutil.parser

# First Party
import alog

# Local
from . import status
from .session import _SESSION_NAMESPACE  # pylint: disable=cyclic-import

## Globals #####################################################################

log = alog.use_channel("VERFY")

DEFAULT_TIMESTAMP_KEY = "lastTransitionTime"
AVAILABLE_CONDITION_KEY = "Available"
COMPLETE_CONDITION_KEY = "Complete"
PROGRESSING_CONDITION_KEY = "Progressing"
NEW_RS_AVAILABLE_REASON = "NewReplicaSetAvailable"

# Type definition for the signature of a resource verify function
# NOTE: I'm not sure why pylint dislikes this name. In my view, this is a shared
#   global which should have all-caps casing.
RESOURCE_VERIFY_FUNCTION = Callable[[dict], bool]  # pylint: disable=invalid-name


## Main Functions ##############################################################


def verify_resource(
    kind: str,
    name: str,
    api_version: str,
    session,
    *,
    # Use a predefined _SESSION_NAMESPACE default instead of None to differentiate between
    # non-namespaced resources (which pass None) and those that use session.namespace
    namespace: Optional[str] = _SESSION_NAMESPACE,
    verify_function: Optional[RESOURCE_VERIFY_FUNCTION] = None,
    is_subsystem: bool = False,
    condition_type: Optional[str] = None,
    timestamp_key: Optional[str] = None,
) -> bool:
    """Verify a resource detailed in a ManagedObject.

    NOTE: we can't do type-hinting on the session because importing
        DeploySession creates a circular dependency with Component. We should
        probably fix that...

    This function will run the appropriate verification function for the given
    resource kind.

    Args:
        kind:  str
            The kind of the resource to look for
        name:  str
            The name of the resource to look for
        api_version:  str
            The api_version of the resource to look for
        session:  DeploySession
            The current deployment session

    Kwargs:
        is_subsystem:  bool
            Whether or not the given kind is an oper8 subsystem
        condition_type:  str
            For non-standard types, this is the type name for the condition to
            check for verification
        timestamp_key:  str
            For non-standard types, this is the key in the condition to use to
            sort by date

    Returns:
        success:  bool
            True on successful deployment verification, False on failure
            conditions
    """

    # Configure namespace if it isn't set
    namespace = namespace if namespace != _SESSION_NAMESPACE else session.namespace

    # Get the state of the object
    log.debug2("Fetching current content for [%s/%s] to verify it", kind, name)
    success, content = session.get_object_current_state(
        kind=kind, name=name, api_version=api_version, namespace=namespace
    )
    assert success, f"Failed to fetch state of [{kind}/{name}]"

    # If the object is not found, it is not verified
    if not content:
        log.debug("Could not find [%s/%s]. Not Ready.", kind, name)
        return False

    # If a custom condition_type is given, then use the general condition
    # verifier
    if condition_type is not None:
        log.debug(
            "Using custom verification for [%s/%s] with condition [%s]",
            kind,
            name,
            condition_type,
        )
        return _verify_condition(
            content, condition_type, True, timestamp_key or DEFAULT_TIMESTAMP_KEY
        )

    # Run the appropriate verification function if there is one available
    verify_fn = verify_function or _resource_verifiers.get(kind)
    if not verify_fn and is_subsystem:
        log.debug("Using oper8 subsystem verifier for [%s/%s]", kind, name)
        verify_fn = partial(
            verify_subsystem,
            desired_version=session.version,
        )

    # If a verifier was found, run it
    if verify_fn:
        log.debug2("Running [%s] verifier for [%s/%s]", kind, kind, name)
        return verify_fn(content)

    # If no other verifier found, we consider it verified as long as it is
    # present in the cluster
    log.debug2("No kind-specific verifier for [%s/%s]", kind, name)
    return True


## Individual Resources ########################################################


def verify_pod(object_state: dict) -> bool:
    """Verify that a pod resources is ready"""
    return _verify_condition(object_state, "Ready", True)


def verify_job(object_state: dict) -> bool:
    """Verify that a job has completed successfully"""
    # https://kubernetes.io/docs/reference/kubernetes-api/workload-resources/job-v1/#JobStatus
    return _verify_condition(object_state, COMPLETE_CONDITION_KEY, True)


def verify_deployment(object_state: dict) -> bool:
    """Verify that all members of a deployment are ready
    and all members are rolled out to new version in case of update.
    """
    return _verify_condition(
        object_state, AVAILABLE_CONDITION_KEY, True
    ) and _verify_condition(
        object_state,
        PROGRESSING_CONDITION_KEY,
        True,
        expected_reason=NEW_RS_AVAILABLE_REASON,
    )


def verify_statefulset(object_state: dict) -> bool:
    """Verify that all desired replicas of a StatefulSet are ready"""
    obj_status = object_state.get("status", {})
    expected_replicas = obj_status.get("replicas")
    if expected_replicas is None:
        log.debug2("No replicas found in statefulset status. Not ready.")
        return False
    ready_replicas = obj_status.get("readyReplicas", 0)
    return ready_replicas == expected_replicas


def verify_subsystem(object_state: dict, desired_version: str = None) -> bool:
    """Verify that an oper8-managed subsystem is ready"""

    current_version = status.get_version(object_state.get("status", {}))
    # Once rollout finishes with verification, version status is added.
    #   Until then, mark the subsystem as unverified.
    if desired_version and not current_version:
        log.debug2(
            "Reconciled version %s does not match desired: %s",
            current_version,
            desired_version,
        )
        return False

    return (
        _verify_condition(
            object_state, status.READY_CONDITION, True, status.TIMESTAMP_KEY
        )
        and _verify_condition(
            object_state, status.UPDATING_CONDITION, False, status.TIMESTAMP_KEY
        )
        and current_version == desired_version
    )


_resource_verifiers = {
    "Pod": verify_pod,
    "Job": verify_job,
    "Deployment": verify_deployment,
    "StatefulSet": verify_statefulset,
}

## Helpers #####################################################################


def _verify_condition(
    object_state: dict,
    type_val: str,
    expected_status: bool,
    timestamp_key: str = DEFAULT_TIMESTAMP_KEY,
    expected_reason: Optional[str] = None,
) -> bool:
    """Perform the guts of checking for an expected condition has expected status
    and reason
    """

    # Look for the condition of the given type
    conditions = _get_conditions(object_state, type_val)
    log.debug2("Found %d [%s] conditions", len(conditions), type_val)

    # If no conditions, the resource is not verified
    if not conditions:
        log.debug2("No %s conditions. Not verified", type_val)
        return False

    # Sort the conditions by transaction time
    latest_cond = _sort_conditions_by_date(conditions, timestamp_key)[0]
    log.debug3("Latest '%s' condition: %s", type_val, latest_cond)
    return _check_condition(latest_cond, expected_status, expected_reason)


def _get_conditions(object_state: dict, type_val: str) -> List[dict]:
    """Get the list of conditions from an object state"""
    return [
        cond
        for cond in object_state.get("status", {}).get("conditions", [])
        if cond.get("type") == type_val
    ]


def _parse_condition_timestamp(condition: dict, timestamp_key: str) -> dict:
    """Parse the timestamp in a condition in place"""
    timestamp = condition.get(timestamp_key)
    log.debug3("Timestamp [%s]: %s", timestamp_key, timestamp)
    if isinstance(timestamp, str):
        return dateutil.parser.parse(timestamp)
    if isinstance(timestamp, datetime):
        return timestamp
    log.warning("Found condition with no valid timestamp. Using epoch")
    return datetime.fromtimestamp(0)


def _sort_conditions_by_date(conditions: List[dict], timestamp_key: str) -> List[dict]:
    """Helper to parse datestamps and sort a list of conditions. The sort will
    put newest conditions first.
    """

    return sorted(
        conditions,
        key=lambda x: _parse_condition_timestamp(x, timestamp_key),
        reverse=True,
    )


def _check_condition(
    condition: dict, expected_status: bool, expected_reason: Optional[str] = None
) -> bool:
    """Helper to parse a condition object and check if it has expected values."""

    def is_expected_status() -> bool:
        """Helper to parse the various ways a 'status' may be represented in a
        condition
        """
        obj_status = condition.get("status")
        if not obj_status:
            return False
        if isinstance(obj_status, str):
            return obj_status.lower() == str(expected_status).lower()
        return bool(obj_status) == expected_status

    def is_expected_reason() -> bool:
        if expected_reason is None:
            return True
        obj_reason = condition.get("reason")
        return obj_reason == expected_reason

    return is_expected_status() and is_expected_reason()



"""
This file holds functions that should be used outside of this module by
components, subsystems, and applications that need access to the TLS context
functionality.
"""

# Standard
from typing import Tuple

# Local
from .factory import get_tls_context, register_tls_context_type
from .internal import InternalTlsContext
from oper8 import Session

## Type Registrations ##########################################################

register_tls_context_type(InternalTlsContext)

## Public Functions ############################################################


def request_server_key_cert_pair(
    session: Session,
    *args,
    **kwargs,
) -> None:
    """Request creation of the PEM encoded value of the key/cert pair for a
    given server. This function has to be called from before render_chart is
    called. I.e., parse_config / Component constructor phase. Implementations of
    this function will generate the pair (in background) if it has not been
    already requested.

    Args:
        session:  Session
            The current deploy session

    Passthrough Args:
        See ITlsContext.request_server_key_cert_pair
    """
    return get_tls_context(session).request_server_key_cert_pair(*args, **kwargs)


def get_server_key_cert_pair(
    session: Session,
    *args,
    **kwargs,
) -> Tuple[str, str]:
    """Get the previously requested PEM encoded value of the key/cert pair for a
    given server. Implementations will retrieveh the pair if it does not exist
    and will fetch its content if it does. If the content is not available, the
    assertion is triggered.

    Args:
        session:  Session
            The current deploy session

    Passthrough Args:
        See ITlsContext.get_server_key_cert_pair

    Returns:
        key_pem:  str
            This is the pem-encoded key content (base64 encoded if
            encode is set)
        cert_pem:  str
            This is the pem-encoded cert content (base64 encoded
            if encode is set)
    """
    return get_tls_context(session).get_server_key_cert_pair(*args, **kwargs)


def get_client_cert(
    session: Session,
    *args,
    **kwargs,
) -> str:
    """Get the CA's public cert

    Args:
        session:  Session
            The current deploy session

    Passthrough Args:
        See ITlsContext.get_client_cert

    Returns:
        crt_pem:  Optional[str]
                           The pem-encoded cert (base64 encoded if encode set),
    """
    return get_tls_context(session).get_client_cert(*args, **kwargs)



"""
Test the patching semantics
"""

# Third Party
import pytest

# First Party
import aconfig
import alog

# Local
from oper8.patch import JSON_PATCH_6902, STRATEGIC_MERGE_PATCH, apply_patches
from oper8.test_helpers.helpers import configure_logging, make_patch

configure_logging()

log = alog.use_channel("TEST")

## Helpers #####################################################################


def sample_foo(base):
    base.setdefault("kind", "Foo")
    base.setdefault("apiVersion", "v1")
    base.setdefault("metadata", {}).setdefault("name", "foo")
    return base


def sample_pod(containers=None):
    return {
        "kind": "Pod",
        "apiVersion": "v1",
        "metadata": {"name": "foo"},
        "spec": {"containers": containers or []},
    }


def pod_psm_body(containers=None, object_name="foo.pod"):
    out = {}
    inner = out
    for part in object_name.split("."):
        inner = inner.setdefault(part, {})
    inner.setdefault("spec", {})["containers"] = containers or []
    return out


## Tests #######################################################################

###################
## patchJson6902 ##
###################


def test_js6902_apply_patch():
    """Test that a simple patchJson6902 patch applies cleanly"""
    obj = sample_foo({"key": "value", "nested": {"key": "nested_value"}})
    patch = make_patch(
        JSON_PATCH_6902,
        {
            "foo": {
                "blob": [
                    {"op": "replace", "path": "/key", "value": "replaced"},
                    {
                        "op": "replace",
                        "path": "/nested/key",
                        "value": "nested_replaced",
                    },
                ]
            }
        },
    )
    res = apply_patches("foo.blob", obj, [patch])
    assert res["key"] == "replaced"
    assert res["nested"]["key"] == "nested_replaced"


def test_js6902_non_list():
    """Test that a non-list js6902 patch raises an error"""
    obj = sample_foo({"key": "value", "nested": {"key": "nested_value"}})
    patch = make_patch(
        JSON_PATCH_6902,
        {
            "foo": {
                "blob": {"op": "replace", "path": "/key", "value": "replaced"},
            }
        },
    )
    with pytest.raises(ValueError):
        apply_patches("foo.blob", obj, [patch])


#########################
## patchStrategicMerge ##
#########################

## Happy Path ##


def test_psm_apply_patch():
    """Test that a simple patchStrategicMerge patch applies cleanly"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "Always"},
                {"name": "bar", "image": "bar"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch])
    assert len(res["spec"]["containers"]) == 2
    assert res["spec"]["containers"][0] == {
        "name": "foo",
        "image": "foo",
        "restartPolicy": "Always",
    }
    assert res["spec"]["containers"][1] == {"name": "bar", "image": "bar"}


def test_psm_update_key():
    """Test that a key can be updated in an object"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "OnFailure"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch])
    assert len(res["spec"]["containers"]) == 1
    assert res["spec"]["containers"][0] == {
        "name": "foo",
        "image": "foo",
        "restartPolicy": "OnFailure",
    }


def test_psm_delete_key():
    """Test that a key can be deleted from an object"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": None},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch])
    assert len(res["spec"]["containers"]) == 1
    assert res["spec"]["containers"][0] == {"name": "foo", "image": "foo"}


def test_psm_delete_dict_list_item():
    """Test that an element can be deleted from a list of dicts"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "$patch": "delete"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch])
    assert len(res["spec"]["containers"]) == 0


def test_psm_delete_primitive_list_items():
    """Test that an element can be deleted from a list of primitives"""
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"foo": {"$deleteFromPrimitiveList/key": [2, 4]}}},
    )
    res = apply_patches("foo.foo", obj, [patch])
    assert res["key"] == [1, 3]


def test_psm_list_element_replace():
    """Test that an element in a list can be replaced without merging"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    replaced_foo = {
        "name": "foo",
        "$patch": "replace",
        "image": "foo:latest",
        "foo": "bar",
    }
    patch = make_patch(STRATEGIC_MERGE_PATCH, pod_psm_body([replaced_foo]))
    res = apply_patches("foo.pod", obj, [patch])
    assert len(res["spec"]["containers"]) == 1
    del replaced_foo["$patch"]
    assert res["spec"]["containers"][0] == replaced_foo


def test_psm_list_no_merge_key_replace():
    """Test that for a list with no merge key, it is direclty replaced"""
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = make_patch(STRATEGIC_MERGE_PATCH, {"foo": {"foo": {"key": [2, 4]}}})
    res = apply_patches("foo.foo", obj, [patch])
    assert res["key"] == [2, 4]


## Error Cases ##


def test_psm_bad_directive():
    """Test that an unknown directive throws an error"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, pod_psm_body([{"name": "foo", "$patch": "flipflop"}])
    )
    with pytest.raises(ValueError):
        apply_patches("foo.pod", obj, [patch])


def test_psm_missing_merge_key_patch():
    """Test that an element in a patch list that is missing the merge key
    causes an error
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = make_patch(STRATEGIC_MERGE_PATCH, pod_psm_body([{"foo": "bar"}]))
    with pytest.raises(ValueError):
        apply_patches("foo.pod", obj, [patch])


def test_psm_missing_merge_key_current():
    """Test that an element in a current list that is missing the merge key
    causes an error
    """
    obj = sample_pod([{"image": "foo"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, pod_psm_body([{"name": "foo", "foo": "bar"}])
    )
    with pytest.raises(ValueError):
        apply_patches("foo.pod", obj, [patch])


def test_psm_primitive_delete_not_there():
    """Test that a request to delete a primitive from a list causes an error if
    the element is not present
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"foo": {"$deleteFromPrimitiveList/keyNotThere": [2, 4]}}},
    )
    with pytest.raises(ValueError):
        apply_patches("foo.foo", obj, [patch])


def test_psm_primitive_delete_patch_not_list():
    """Test that a request to delete a primitive from a list where the patch
    value is not a list causes an error
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, {"foo": {"foo": {"$deleteFromPrimitiveList/key": 2}}}
    )
    with pytest.raises(ValueError):
        apply_patches("foo.foo", obj, [patch])


def test_psm_primitive_delete_current_not_list():
    """Test that a request to delete a primitive from a list where the current
    value is not a list causes an error
    """
    obj = sample_foo({"key": {"a": 1, "b": 2}})
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, {"foo": {"foo": {"$deleteFromPrimitiveList/key": ["a"]}}}
    )
    with pytest.raises(ValueError):
        apply_patches("foo.foo", obj, [patch])


def test_psm_primitive_delete_element_not_found():
    """Test that a request to delete a primitive from a list which doesn't
    contain the element raises an error
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, {"foo": {"foo": {"$deleteFromPrimitiveList/key": [5]}}}
    )
    with pytest.raises(ValueError):
        apply_patches("foo.foo", obj, [patch])


def test_psm_delete_element_not_found():
    """Test that a `$patch: delete` directive to delete an element from a list
    that isn't found raises an error
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = make_patch(
        STRATEGIC_MERGE_PATCH, pod_psm_body([{"name": "bar", "$patch": "delete"}])
    )
    with pytest.raises(ValueError):
        apply_patches("foo.pod", obj, [patch])


#############
## general ##
#############


def test_bad_patch_type():
    """Test that a bad patch type raises an error"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    with pytest.raises(ValueError):
        apply_patches(
            "foo.pod", obj, [make_patch("BAD_TYPE", pod_psm_body([{"name": "foo"}]))]
        )


def test_mixed_patch_types():
    """Test that multiple patches of different types can be applied to the same
    object
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch1 = make_patch(
        JSON_PATCH_6902,
        {
            "foo": {
                "pod": [
                    {"op": "add", "path": "/spec/imagePullPolicy", "value": "Always"}
                ]
            }
        },
    )
    patch2 = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "Always"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch1, patch2])
    expected = obj
    expected["spec"]["imagePullPolicy"] = "Always"
    expected["spec"]["containers"][0]["restartPolicy"] = "Always"
    assert res == expected


def test_patch_order():
    """Test that an ordered list of patches is applied in the correct order"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch1 = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "Always", "image": "image1"},
            ]
        ),
    )
    patch2 = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "OnFailure"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch1, patch2])
    expected = obj
    expected["spec"]["containers"][0]["restartPolicy"] = "OnFailure"
    expected["spec"]["containers"][0]["image"] = "image1"
    assert res == expected


def test_only_applicable_patches():
    """Test that only the patches for the given object are applied, and all
    other patches are ignored
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch1 = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "Always", "image": "image1"},
            ],
            object_name="bar.pod",
        ),
    )
    patch2 = make_patch(
        STRATEGIC_MERGE_PATCH,
        pod_psm_body(
            [
                {"name": "foo", "restartPolicy": "OnFailure"},
            ]
        ),
    )
    res = apply_patches("foo.pod", obj, [patch1, patch2])
    expected = obj
    expected["spec"]["containers"][0]["restartPolicy"] = "OnFailure"
    assert res == expected
    # Just to make it clear
    assert res["spec"]["containers"][0]["image"] == "foo"



"""
Tests for common PWM utils
"""
# Standard
from datetime import timedelta
from unittest import mock
from uuid import uuid4
import logging

# Third Party
import pytest

# Local
from oper8.test_helpers.helpers import library_config
from oper8.watch_manager.python_watch_manager.utils import common


@pytest.mark.parametrize(
    "test_cfg",
    [
        ("12hr13m14s", timedelta(hours=12, minutes=13, seconds=14)),
        ("0hr0m62s", timedelta(minutes=1, seconds=2)),
        ("0hr0m62.5s", timedelta(minutes=1, seconds=2.5)),
        ("10s", timedelta(seconds=10)),
        ("foobar", None),
        ("0hr0m1.2.3s", None),
    ],
)
def test_parse_time_delta(test_cfg):
    """Make sure that the parser behaves as expected for valid and invalid
    timedeltas
    """
    time_str, result = test_cfg
    assert common.parse_time_delta(time_str) == result


def test_get_operator_namespace_from_file():
    ns_name = "some-namespace"
    other_ns_name = "other-namespace"
    path_mock = mock.MagicMock()
    path_mock.is_file = mock.MagicMock(return_value=True)
    path_mock.read_text = mock.MagicMock(return_value=ns_name)
    with mock.patch("pathlib.Path", return_value=path_mock), library_config(
        python_watch_manager={"lock": {"namespace": other_ns_name}}
    ):
        assert common.get_operator_namespace() == ns_name
        assert path_mock.is_file.called
        assert path_mock.read_text.called


def test_get_operator_namespace_from_config():
    ns_name = "some-namespace"
    other_ns_name = "other-namespace"
    path_mock = mock.MagicMock()
    path_mock.is_file = mock.MagicMock(return_value=False)
    path_mock.read_text = mock.MagicMock(return_value=ns_name)
    with mock.patch("pathlib.Path", return_value=path_mock), library_config(
        python_watch_manager={"lock": {"namespace": other_ns_name}}
    ):
        assert common.get_operator_namespace() == other_ns_name
        assert path_mock.is_file.called
        assert not path_mock.read_text.called


def test_get_logging_handler_adds_stream_handlers():
    """Make sure that get_logging_handlers adds a stream handler by default if
    no other handlers configured
    """
    logger = logging.Logger(str(uuid4()))
    with mock.patch("logging.getLogger", return_value=logger):
        returned_handlers = common.get_logging_handlers()
    assert len(returned_handlers) == 1
    assert isinstance(returned_handlers[0], logging.StreamHandler)


def test_get_logging_handler_does_not_overwrite_other_handlers():
    """Make sure that get_logging_handlers does not change preconfigured
    handlers
    """
    logger = logging.Logger(str(uuid4()))
    logger.addHandler(logging.NullHandler())
    with mock.patch("logging.getLogger", return_value=logger):
        returned_handlers = common.get_logging_handlers()
    assert len(returned_handlers) == 1
    assert isinstance(returned_handlers[0], logging.NullHandler)



"""
This module holds common functionality that the DeployManager implementations
can use to manage ownerReferences on deployed resources
"""

# First Party
import alog

# Local
from ..exceptions import assert_cluster
from .base import DeployManagerBase

log = alog.use_channel("OWNRF")


def update_owner_references(
    deploy_manager: DeployManagerBase,
    owner_cr: dict,
    child_obj: dict,
):
    """Fetch current ownerReferences and merge a reference for this CR into
    the child object
    """

    # Validate the shape of the owner CR and the chid object
    _validate_object_struct(owner_cr)
    _validate_object_struct(child_obj)

    # Fetch the current state of this object
    kind = child_obj["kind"]
    api_version = child_obj["apiVersion"]
    name = child_obj["metadata"]["name"]
    namespace = child_obj["metadata"]["namespace"]
    uid = child_obj["metadata"].get("uid")

    success, content = deploy_manager.get_object_current_state(
        kind=kind, name=name, api_version=api_version, namespace=namespace
    )
    assert_cluster(
        success, f"Failed to fetch current state of {api_version}.{kind}/{name}"
    )

    # Get the current ownerReferences
    owner_refs = []
    if content is not None:
        owner_refs = content.get("metadata", {}).get("ownerReferences", [])
        log.debug3("Current owner refs: %s", owner_refs)

    # If the current CR is not represented and current CR is in the same
    # namespace as the child object, add it
    current_uid = owner_cr["metadata"]["uid"]
    log.debug3("Current CR UID: %s", current_uid)
    current_namespace = owner_cr["metadata"]["namespace"]
    log.debug3("Current CR namespace: %s", current_namespace)

    if current_uid == uid:
        log.debug2("Owner is same as child; Not adding owner ref")
        return

    if (namespace == current_namespace) and (
        current_uid not in [ref["uid"] for ref in owner_refs]
    ):
        log.debug2(
            "Adding current CR owner reference for %s.%s/%s",
            api_version,
            kind,
            name,
        )
        owner_refs.append(_make_owner_reference(owner_cr))

    # Add the ownerReferences to the object that will be applied to the
    # cluster
    log.debug4("Final owner refs: %s", owner_refs)
    child_obj["metadata"]["ownerReferences"] = owner_refs


## Implementation Details ######################################################


def _validate_object_struct(obj: dict):
    """Ensure that the required portions of an object are present (kind,
    apiVerison, metadata.namespace, metadata.name)
    """
    assert "kind" in obj, "Got object without 'kind'"
    assert "apiVersion" in obj, "Got object without 'apiVersion'"
    metadata = obj.get("metadata")
    assert isinstance(metadata, dict), "Got object with non-dict 'metadata'"
    assert "name" in metadata, "Got object without 'metadata.name'"
    assert "namespace" in metadata, "Got object without 'metadata.namespace'"


def _make_owner_reference(owner_cr: dict) -> dict:
    """Make an owner reference for the given CR instance

    Error Semantics: This function makes a best-effort and does not validate the
    content of the owner_cr, so the resulting ownerReference may contain None
    entries.

    Args:
        owner_cr:  dict
            The full CR manifest for the owning resource

    Returns:
        owner_reference:  dict
            The dict entry for the `metadata.ownerReferences` entry of the owned
            object
    """
    # NOTE: We explicitly don't set controller: True here. If two
    #   oper8-managed resources reference the resource, only one can have
    #   controller set to True. According to StackOverflow, this field is
    #   only used for adoption and not garbage collection.
    # CITE: https://stackoverflow.com/a/65825463
    metadata = owner_cr.get("metadata", {})
    return {
        "apiVersion": owner_cr.get("apiVersion"),
        "kind": owner_cr.get("kind"),
        "name": metadata.get("name"),
        "uid": metadata.get("uid"),
        # The parent will not be deleted until this object completes its
        # deletion
        "blockOwnerDeletion": True,
    }



"""
Base operator config module. The config here is only used as a baseline bootup
config. All application-specific config must come from the app_config.
"""

# Standard
import sys as _sys

# Local
from .config import library_config


# Define __getattr__ on this module to delegate to the library config.
def __getattr__(name):
    if name in library_config or hasattr({}, name):
        return getattr(library_config, name)
    raise AttributeError(f"No such config attribute {name}")


# Only expose the library config keys
__all__ = list(library_config.keys())



"""
This module contains a collection of classes for implementing nodes of a Graph
"""
# Standard
from typing import Any, Callable, List, Optional


class Node:
    """Class for representing a node in the Graph"""

    def __init__(
        self,
        name: Optional[str] = None,
        data: Optional[Any] = None,
    ) -> None:
        """Construct a new Node

        Args:
            name:  Optional[str]
                The name of the node
            data:  Optional[Any]
                Any data that should be stored with the node
        """
        self._name = name
        self._data = data
        self.children = {}

    ## Modifiers ##############################################################
    def add_child(self, node: "Node", edge_data: Optional[Any] = None):
        """Add edge from  self to node with optional edge data"""
        if node.dfs(self):
            raise ValueError("Unable to add cyclic dependency")
        self.children[node] = edge_data

    def remove_child(self, node: "Node"):
        """Remove child node from self"""
        if node in self.children:
            self.children.pop(node)

    def set_data(self, data: Any):
        """Mutator for node data"""
        self._data = data

    ## Accessors ##############################################################

    def get_data(self):
        """Accessor for specific child"""
        return self._data

    def get_name(self):
        """Accessor for specific child"""
        return self._name

    def has_child(self, node: "Node"):
        """Accessor for specific child"""
        return node in self.children

    def get_children(self) -> set:
        """Accessor for all children"""
        return list(self.children.items())

    ## Graph Functions ##############################################################
    def topology(self) -> List["Node"]:
        """Function to get an ordered topology of a node's children"""
        found = set()
        topology = []

        def visit(node):
            for child, _ in sorted(node.get_children()):
                visit(child)

            if node not in found:
                topology.append(node)
                found.add(node)

        visit(self)

        return topology

    def dfs(self, node: "Node", visited: List["Node"] = None) -> bool:
        """Function to determine if their is a path between two nodes. Used in acyclic check"""
        if not visited:
            visited = []
        if node == self:
            return True
        visited.append(self)
        for child, _ in self.get_children():
            if child not in visited:
                if child.dfs(node, visited):
                    return True
                visited.append(child)
        return False

    ## Internal ##
    def __eq__(self, obj):
        """Compare and sort nodes by name"""
        if not isinstance(obj, Node):
            return False
        return (self.get_name()) == (obj.get_name())

    def __lt__(self, obj):
        if not isinstance(obj, Node):
            return False
        return (self.get_name()) < (obj.get_name())

    def __repr__(self) -> str:
        # __repr__ may be called before __init__ thus _name is not present
        if hasattr(self, "_name"):
            return f"{self.__class__.__name__}('{self.get_name()}', {self.get_data()})"
        return super().__repr__()

    def __hash__(self) -> str:
        return self.get_name().__hash__()


class ResourceNode(Node):
    """Class for representing a kubernetes resource in the Graph with
    a function for verifying said resource"""

    def __init__(
        self,
        name: str,
        manifest: dict,
        verify_func: Optional[Callable] = None,
        deploy_method: Optional["DeployMethod"] = None,  # noqa: F821
    ):
        # Override init to require name/manifest parameters
        super().__init__(name, manifest)
        self._verify_function = verify_func
        self._deploy_method = deploy_method
        if not deploy_method:
            # Local
            from ..deploy_manager import DeployMethod

            self._deploy_method = DeployMethod.DEFAULT

    ## ApiObject Parameters and Functions ######################################
    @property
    def api_group(self) -> str:
        """The kubernetes apiVersion group name without the schema version"""
        return self.api_version.split("/")[0]

    @property
    def api_version(self) -> str:
        """The full kubernetes apiVersion"""
        return self.manifest.get("apiVersion")

    @property
    def kind(self) -> str:
        """The resource kind"""
        return self.manifest.get("kind")

    @property
    def metadata(self) -> dict:
        """The full resource metadata dict"""
        return self.manifest.get("metadata", {})

    @property
    def name(self) -> str:
        """The resource metadata.name"""
        return self.metadata.get("name")

    @property
    def manifest(self) -> dict:
        """The resource manifest"""
        return self.get_data()

    @property
    def verify_function(self) -> Optional[Callable]:
        """The resource manifest"""
        return self._verify_function

    @property
    def deploy_method(self) -> Optional["DeployMethod"]:  # noqa: F821
        """The resource manifest"""
        return self._deploy_method

    def add_dependency(self, node: "ResourceNode"):
        """Add a child dependency to this node"""
        self.add_child(node)



"""
This is the main entrypoint command for running the operator
"""
# Standard
from datetime import datetime, timedelta
from pathlib import Path
import argparse

# First Party
import alog

# Local
from .. import config
from ..watch_manager.python_watch_manager.threads.heartbeat import HeartbeatThread
from .base import CmdBase

log = alog.use_channel("MAIN")


class CheckHeartbeatCmd(CmdBase):
    __doc__ = __doc__

    ## Interface ##

    def add_subparser(
        self,
        subparsers: argparse._SubParsersAction,
    ) -> argparse.ArgumentParser:
        parser = subparsers.add_parser("check-heartbeat", help=__doc__)
        runtime_args = parser.add_argument_group("Check Heartbeat Configuration")
        runtime_args.add_argument(
            "--delta",
            "-d",
            required=True,
            type=int,
            help="Max time allowed since last check",
        )
        runtime_args.add_argument(
            "--file",
            "-f",
            default=config.python_watch_manager.heartbeat_file,
            help="Location of health check file. Defaults to config based.",
        )
        return parser

    def cmd(self, args: argparse.Namespace):
        """Run command to validate a health check file"""

        # Validate args
        assert args.delta is not None
        assert args.file is not None

        # Ensure file exists
        file_path = Path(args.file)
        if not file_path.exists():
            log.error(f"Health Check failed: {file_path} does not exist")
            raise FileNotFoundError()

        # Read and the most recent time from the health check
        last_log_time = file_path.read_text().strip()
        last_time = datetime.strptime(last_log_time, HeartbeatThread._DATE_FORMAT)

        if last_time + timedelta(seconds=args.delta) < datetime.now():
            msg = f"Health Check failed: {last_log_time} is to old"
            log.error(msg)
            raise KeyError(msg)



"""
Test the implementation of the log_rotator open_iteration function
"""

# Standard
import json
import os
import shlex
import subprocess
import sys
import tempfile

# First Party
import alog

# Local
from oper8.test_helpers.helpers import configure_logging

configure_logging()
log = alog.use_channel("TEST")


def strip_indentation(script_string):
    # Strip off "extra" indentation. This allows the tests to not look TERRIBLE
    lines = script_string.split("\n")
    first_line = [l for l in lines if l.strip()][0]
    indentation = first_line[: len(first_line) - len(first_line.lstrip())]
    if all([l.startswith(indentation) or not l for l in lines]):
        lines = [l[len(indentation) :] for l in lines]
    return "\n".join(lines)


def run_test_script(script_string):
    """Since the goal of the rotator is to perform log rotation over a series of
    successive python process executions, we use this helper to launch
    subprocesses
    """

    script = strip_indentation(script_string)
    with tempfile.NamedTemporaryFile(mode="w") as f_handle:
        # Save the script
        f_handle.write(script)
        f_handle.flush()

        # Run it
        assert os.path.exists(f_handle.name)
        cmd = f"{sys.executable} {f_handle.name}"
        with alog.ContextTimer(log.info, "Subprocess runtime: "):
            subprocess.run(shlex.split(cmd))


def parse_log_files(workdir):
    out = {}
    for fname in os.listdir(workdir):
        fpath = os.path.join(workdir, fname)
        with open(fpath, "r") as log_handle:
            out[fname] = json.load(log_handle)
    return out


def test_alog_configure():
    """Test that a log file is created when using the AutoRotatingFileHandler
    with alog.configure
    """
    with tempfile.TemporaryDirectory() as workdir:
        log_file_name = "foo.log"

        # Define the "doit" script which will open the log file for rotation
        def get_script(i):
            return f"""
                import alog
                from oper8.watch_manager.ansible_watch_manager.modules.log_rotator import AutoRotatingFileHandler
                alog.configure(
                    "info",
                    formatter="json",
                    handler_generator=lambda: AutoRotatingFileHandler(
                        '{os.path.join(workdir, log_file_name)}',
                        backupCount=1,
                    ),
                )
                ch = alog.use_channel("TEST")
                ch.info("TEST {i}")
                """

        # Run the script the first time
        run_test_script(get_script(1))
        all_files = parse_log_files(workdir)
        assert len(all_files) == 1
        assert log_file_name in all_files
        assert all_files[log_file_name]["message"] == "TEST 1"

        # Run it a second time and make sure it keeps one backup
        run_test_script(get_script(2))
        all_files = parse_log_files(workdir)
        assert len(all_files) == 2
        assert log_file_name in all_files
        assert all_files[log_file_name]["message"] == "TEST 2"
        assert all_files[f"{log_file_name}.1"]["message"] == "TEST 1"

        # Run it a third time and make sure it keeps only one backup
        run_test_script(get_script(3))
        all_files = parse_log_files(workdir)
        assert len(all_files) == 2
        assert log_file_name in all_files
        assert all_files[log_file_name]["message"] == "TEST 3"
        assert all_files[f"{log_file_name}.1"]["message"] == "TEST 2"



"""
Helper object to represent a kubernetes object that is managed by the operator
"""
# Standard
from typing import Callable, Optional
import uuid

KUBE_LIST_IDENTIFIER = "List"


class ManagedObject:  # pylint: disable=too-many-instance-attributes
    """Basic struct to represent a managed kubernetes object"""

    def __init__(
        self,
        definition: dict,
        verify_function: Optional[Callable] = None,
        deploy_method: Optional["DeployMethod"] = None,  # noqa: F821
    ):
        self.kind = definition.get("kind")
        self.metadata = definition.get("metadata", {})
        self.name = self.metadata.get("name")
        self.namespace = self.metadata.get("namespace")
        self.uid = self.metadata.get("uid", uuid.uuid4())
        self.resource_version = self.metadata.get("resourceVersion")
        self.api_version = definition.get("apiVersion")
        self.definition = definition
        self.verify_function = verify_function
        self.deploy_method = deploy_method

        # If resource is not list then check name
        if KUBE_LIST_IDENTIFIER not in self.kind:
            assert self.name is not None, "No name found"

        assert self.kind is not None, "No kind found"
        assert self.api_version is not None, "No apiVersion found"

    def get(self, *args, **kwargs):
        """Pass get calls to the objects definition"""
        return self.definition.get(*args, **kwargs)

    def __str__(self):
        return f"{self.api_version}/{self.kind}/{self.name}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        """Hash explicitly excludes the definition so that the object's
        identifier in a map can be based only on the unique identifier of the
        resource in the cluster. If the original resource did not provide a unique
        identifier then use the apiVersion, kind, and name
        """
        return hash(self.metadata.get("uid", str(self)))

    def __eq__(self, other):
        return hash(self) == hash(other)



"""
Shared test config
"""
# Standard
from unittest import mock
import sys

# Third Party
import pytest

# Local
from oper8.reconcile import ReconcileManager
from oper8.test_helpers.helpers import (
    config_detail_dict,
    configure_logging,
    version_safe_md5,
)

configure_logging()


@pytest.fixture(autouse=True)
def no_local_kubeconfig():
    """This fixture makes sure the tests run as if KUBECONFIG is not exported in
    the environment, even if it is
    """
    with mock.patch(
        "kubernetes.config.new_client_from_config", side_effect=RuntimeError
    ):
        yield


@pytest.fixture(autouse=True)
def fork_multiprocess():
    """This fixture makes sure that we use fork for multiprocessing instead of
    spawn. Spawn is more reliable with FIPs and OpenSSL but causes issues when
    pickling mocked or patched objects.
    """
    # Don't use library_config since some tests read the config object directly
    config_detail_dict.python_watch_manager.process_context = "fork"


@pytest.fixture(autouse=True)
def no_unimport_oper8_mods():
    """Since our helper classes themselves live within the top of oper8, the
    real logic for unimporting the controller modules will recursively reimport
    _all_ oper8.* modules. This causes a lot of problems with tests like
    misbehaving mocks, Controller is not Controller, etc...
    """
    real_unimport = ReconcileManager._unimport_controller_module

    @staticmethod
    def _patched_unimport(module_name):
        oper8_mods = {
            mod_name: mod
            for mod_name, mod in sys.modules.items()
            if mod_name.startswith("oper8.")
            and not mod_name.startswith("oper8.test_helpers")
        }
        reimport_modules = real_unimport(module_name)
        for mod_name, mod in oper8_mods.items():
            sys.modules.setdefault(mod_name, mod)
        return reimport_modules

    with mock.patch(
        "oper8.reconcile.ReconcileManager._unimport_controller_module",
        new=_patched_unimport,
    ):
        yield



"""
Test the TLS context utility functionality
"""

# Standard
import re

# First Party
import aconfig
import alog

# Local
from oper8.test_helpers.helpers import configure_logging, setup_session
from oper8.x.utils.tls_context import factory, internal
from tests.x.utils.tls_context.util import make_server_component_class

configure_logging()
log = alog.use_channel("TEST")

NAME_CLEANER = re.compile("[^-a-zA-Z0-9]")

## Happy Path ##


def test_register_same_cert_twice():
    """Test that it is possible to call register Cert twice (the second call prints WARNING)"""
    tls_cfg = aconfig.Config({"type": "internal"})
    session = setup_session(app_config={"tls": tls_cfg})
    context = internal.InternalTlsContext(session, config=tls_cfg)
    server_comp = make_server_component_class(request_cert=False)(session)
    context.request_server_key_cert_pair(server_comp, ["localhost"])
    context.request_server_key_cert_pair(server_comp, ["localhost"])


def test_passthough_provided_certs():
    """Test that existing certs parameters are passed through"""
    tls_cfg = aconfig.Config({"type": "internal"})
    session = setup_session(deploy_config={"tls": tls_cfg})
    existing_key_pem = "-----BEGIN FAKED PRIVATE KEY-----\nfake-key"
    existing_cert_pem = "-----BEGIN CERTIFICATE-----\nfake-cert"

    context = internal.InternalTlsContext(session, config=tls_cfg)
    server_comp = make_server_component_class(request_cert=False)(session)
    context.request_server_key_cert_pair(server_comp, ["localhost"])

    (key_pem, cert_pem) = context.get_server_key_cert_pair(
        server_component=server_comp,
        encode=False,
        existing_key_pem=existing_key_pem,
        existing_cert_pem=existing_cert_pem,
    )

    assert key_pem == existing_key_pem
    assert cert_pem == existing_cert_pem


def test_multi_session():
    """Test that reusing an existing session where the component has already
    been registered does not error out. This can happen when running in
    standalone mode (i.e. make run)
    """
    session1 = setup_session(app_config={"tls": {"type": "internal"}})

    # Create the instance
    factory.get_tls_context(session1)

    # Create a nested session to simulate standalone recursion
    session2 = setup_session(app_config={"tls": {"type": "internal"}})
    factory.get_tls_context(session2)

    # Use the factory again with the original session
    factory.get_tls_context(session1)


def test_label_overrides():
    """Test that the labels can be overridden in the config"""
    tls_cfg = aconfig.Config(
        {
            "type": "internal",
            "labels": {"foo": "bar"},
        }
    )
    session = setup_session(app_config={"tls": tls_cfg})
    context = internal.InternalTlsContext(session, config=tls_cfg)
    comp = context._component.to_dict(session)[0]
    assert comp["metadata"]["labels"] == tls_cfg.labels



"""
Top-level watch_manager imports
"""

# Local
from .ansible_watch_manager import AnsibleWatchManager
from .base import WatchManagerBase
from .dry_run_watch_manager import DryRunWatchManager
from .python_watch_manager import PythonWatchManager

# Expose the static start/stop functions at the top
start_all = WatchManagerBase.start_all
stop_all = WatchManagerBase.stop_all



"""
This module holds the core session state for an individual reconciliation
"""

# Standard
from functools import partial
from typing import Callable, List, Optional, Tuple, Union
import hashlib

# First Party
import aconfig
import alog

# Local
from .dag import Graph
from .deploy_manager import DeployManagerBase
from .exceptions import assert_cluster
from .status import get_version
from .utils import get_manifest_version

log = alog.use_channel("SESSION")

HASH_CHARSET = "0123456789abcdef"


# Maximum length for a kubernetes name
MAX_NAME_LEN = 63

# Type definition for the signature of a component verify function
# NOTE: I'm not sure why pylint dislikes this name. In my view, this is a shared
#   global which should have all-caps casing.
COMPONENT_VERIFY_FUNCTION = Callable[["Session"], bool]  # pylint: disable=invalid-name

# Helper Definition to define when a session should use its own namespace
# or the one passed in as an argument
_SESSION_NAMESPACE = "__SESSION_NAMESPACE__"

# Forward declaration for Component
COMPONENT_INSTANCE_TYPE = "Component"


class Session:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """A session is the core context manager for the state of an in-progress
    reconciliation
    """

    # We strictly define the set of attributes that a Session can have to
    # disallow arbitrary assignment
    __slots__ = [
        "__components",
        "__component_dependencies",
        "__enabled_components",
        "__disabled_components",
        "__id",
        "__cr_manifest",
        "__config",
        "__temporary_patches",
        "__deploy_manager",
        "__status",
        "__current_version",
        "__graph",
        # _app is retained for backwards compatibility
        "_app",
    ]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        reconciliation_id: str,
        cr_manifest: aconfig.Config,
        config: aconfig.Config,
        deploy_manager: DeployManagerBase,
        temporary_patches: Optional[List[dict]] = None,
    ):
        """Construct a session object to hold the state for a reconciliation

        Args:
            reconciliation_id:  str
                The unique ID for this reconciliation
            cr_manifest:  aconfig.Config
                The full value of the CR manifest that triggered this
                reconciliation
            config:  aconfig.Config
                The compiled backend config for this reconciliation
            deploy_manager:  DeployManagerBase
                The preconfigured DeployManager in charge of running the actual
                deploy operations for this deployment
            temporary_patches:  list(dict)
                List of temporary patch object to apply to resources managed by
                this rollout
        """

        ##################################################################
        # Private Members: These members will be hidden from client code #
        ##################################################################

        # Mapping from component name to Component instance
        self.__graph = Graph()

        ###################################################
        # Properties: These properties will be exposed as #
        # @property members to be used by client code     #
        ###################################################

        self.__id = reconciliation_id
        if not isinstance(cr_manifest, aconfig.Config):
            cr_manifest = aconfig.Config(cr_manifest, override_env_vars=False)
        self._validate_cr(cr_manifest)
        self.__cr_manifest = cr_manifest
        if not isinstance(config, aconfig.Config):
            config = aconfig.Config(config, override_env_vars=False)
        self.__config = config
        self.__temporary_patches = temporary_patches or []

        # The deploy manager that will be used to manage interactions with the
        # cluster
        self.__deploy_manager = deploy_manager

        # Get the current status and version so that it can be referenced by the
        # Application and Components that use it
        self.__status = self.get_status()
        self.__current_version = get_version(self.status)

    ## Properties ##############################################################

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """The unique reconciliation ID"""
        return self.__id

    @property
    def cr_manifest(self) -> aconfig.Config:
        """The full CR manifest that triggered this reconciliation"""
        return self.__cr_manifest

    @property
    def spec(self) -> aconfig.Config:
        """The spec section of the CR manifest"""
        return self.cr_manifest.get("spec", aconfig.Config({}))

    @property
    def version(self) -> str:
        """The spec.version for this CR"""
        return get_manifest_version(self.cr_manifest)

    @property
    def metadata(self) -> aconfig.Config:
        """The metadata for this CR"""
        return self.cr_manifest.metadata

    @property
    def kind(self) -> str:
        """The kind of the operand for this CR"""
        return self.cr_manifest.kind

    @property
    def api_version(self) -> str:
        """The api version of the operand for this CR"""
        return self.cr_manifest.apiVersion

    @property
    def name(self) -> str:
        """The metadata.name for this CR"""
        return self.metadata.name

    @property
    def namespace(self) -> str:
        """The metadata.namespace for this CR"""
        return self.metadata.namespace

    @property
    def finalizers(self) -> str:
        """The metadata.namespace for this CR"""

        # Manually create finalizer list if it doesn't exist so its
        # editable
        if "finalizers" not in self.metadata:
            self.metadata["finalizers"] = []

        return self.metadata.get("finalizers")

    @property
    def config(self) -> aconfig.Config:
        """The backend config for this reconciliation"""
        return self.__config

    @property
    def temporary_patches(self) -> List[aconfig.Config]:
        """Ordered list of temporary patches that apply to the operand being
        reconciled
        """
        return self.__temporary_patches

    @property
    def status(self) -> aconfig.Config:
        """The operand status"""
        return self.__status

    @property
    def current_version(self) -> aconfig.Config:
        """The most recently reconciled version of the operand"""
        return self.__current_version

    @property
    def deploy_manager(self) -> DeployManagerBase:
        """Allow read access to the deploy manager"""
        return self.__deploy_manager

    @property
    def graph(self) -> str:  # pylint: disable=invalid-name
        """The component graph"""
        return self.__graph

    ## State Management ########################################################
    #
    # These functions are used by derived controllers in their setup_components
    # implementations
    ##

    @alog.logged_function(log.debug2)
    def add_component(self, component: COMPONENT_INSTANCE_TYPE):
        """Add a component to this deploy associated with a specific application

        Args:
            component:  Component
                The component to add to this deploy
            disabled:  bool
                Whether or not the component is disabled in this deploy
        """
        self.graph.add_node(component)

    def add_component_dependency(
        self,
        component: Union[str, COMPONENT_INSTANCE_TYPE],
        upstream_component: Union[str, COMPONENT_INSTANCE_TYPE],
        verify_function: Optional[COMPONENT_VERIFY_FUNCTION] = None,
    ):
        """Add a dependency indicating that one component requires an upstream
        component to be deployed before it can be deployed.

        Args:
            component:  str or Component
                The component or name of component in the deploy that must wait for the upstream
            upstream_component:  str or Component
                The upstream component or name of upstream that must be deployed before component
            verify_function:  callable
                A callable function of the form `def verify(session) -> bool:`
                to use to verify that the dependency has been satisfied. This
                will be used to block deployment of the component beyond
                requiring that the upstream has been deployed successfully.
        """
        # Get component obj if name was provided
        component_node = component
        if isinstance(component, str):
            component_node = self.get_component(component)

        upstream_component_node = upstream_component
        if isinstance(upstream_component, str):
            upstream_component_node = self.get_component(upstream_component)

        if not component_node or not upstream_component_node:
            raise ValueError(
                f"Cannot add dependency [{component} -> {upstream_component}]",
                " for unknown component(s)",
            )

        if component_node.disabled or upstream_component_node.disabled:
            raise ValueError(
                f"Cannot add dependency [{component} -> {upstream_component}]",
                " for with disabled component(s)",
            )

        # Add session parameter to verify function if one was provided
        if verify_function:
            verify_function = partial(verify_function, self)
        self.graph.add_node_dependency(
            component_node, upstream_component_node, verify_function
        )

    ## Utilities ###############################################################
    #
    # These utilities may be used anywhere in client code to perform common
    # operations based on the state of the session.
    ##
    def get_component(
        self, name: str, disabled: Optional[bool] = None
    ) -> Optional[COMPONENT_INSTANCE_TYPE]:
        """Get an individual component by name

        Args:
            name: str
                Name of component to return
            disabled: Optional[bool]
                Option on wether to return disabled components. If this option is not supplied then
                the referenced component will be returned irregardless whether its disabled
                or enabled

        Returns:
            component: Optional[Component]
                The component with the given name or None if component does not exit or does
                not match disabled arg
        """
        comp = self.graph.get_node(name)

        # Only filter disabled/enabled components if the option was passed in.
        if isinstance(disabled, bool):
            if disabled:
                return comp if comp.disabled else None
            return comp if not comp.disabled else None

        return comp

    def get_components(self, disabled: bool = False) -> List[COMPONENT_INSTANCE_TYPE]:
        """Get all components associated with an application

        Args:
            disabled:  bool
                Whether to return disabled or enabled components

        Returns:
            components:  list(Component)
                The list of Component objects associated with the given
                application
        """
        assert isinstance(
            disabled, bool
        ), "Disabled flag must be a bool. You may be using the old function signature!"

        # Get list of all components.
        comp_list = self.graph.get_all_nodes()

        # Filter out disabled/enabled components using get_component
        filtered_list = [
            comp for comp in comp_list if self.get_component(comp.get_name(), disabled)
        ]

        return filtered_list

    def get_component_dependencies(
        self,
        component: Union[str, COMPONENT_INSTANCE_TYPE],
    ) -> List[Tuple[COMPONENT_INSTANCE_TYPE, Optional[COMPONENT_VERIFY_FUNCTION]]]:
        """Get the list of (upstream_name, verify_function) tuples for a given
        component.

        NOTE: This is primarily for use inside of the RolloutManager. Do not use
            this method in user code unless you know what you're doing!

        Args:
            component_name:  str
                The name of the component to lookup dependencies for

        Returns:
            upstreams:  List[Tuple[str, Optional[VERIFY_FUNCTION]]]
                The list of upstream (name, verify_fn) pairs
        """
        component_node = component
        if isinstance(component, str):
            component_node = self.get_component(component)

        return component_node.get_children()

    def get_scoped_name(self, name: str) -> str:
        """Get a name that is scoped to the application instance

        Args:
            name:  str
                The name of a resource that will be managed by this operator
                which should have instance name scoping applied

        Returns:
            scoped_name:  str
                The scoped and truncated version of the input name
        """
        scoped_name = self.get_truncated_name(f"{self.name}-{name}")
        log.debug3("Scoped name [%s] -> [%s]", name, scoped_name)
        return scoped_name

    @staticmethod
    def get_truncated_name(name: str) -> str:
        """Perform truncation on a cluster name to make it conform to kubernetes
        limits while remaining unique.

        Args:
            name:  str
                The name of the resource that should be truncated and made
                unique

        Returns:
            truncated_name:  str
                A version of name that has been truncated and made unique
        """
        if len(name) > MAX_NAME_LEN:
            sha = hashlib.sha256()
            sha.update(name.encode("utf-8"))
            trunc_name = name[: MAX_NAME_LEN - 4] + sha.hexdigest()[:4]
            log.debug2("Truncated name [%s] -> [%s]", name, trunc_name)
            name = trunc_name
        return name

    def get_object_current_state(
        self,
        kind: str,
        name: str,
        api_version: Optional[str] = None,
        namespace: Optional[str] = _SESSION_NAMESPACE,
    ) -> Tuple[bool, Optional[dict]]:
        """Get the current state of the given object in the namespace of this
        session

        Args:
            kind:  str
                The kind of the object to fetch
            name:  str
                The full name of the object to fetch
            api_version:  str
                The api_version of the resource kind to fetch

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  dict or None
                The dict representation of the current object's configuration,
                or None if not present
        """
        namespace = namespace if namespace != _SESSION_NAMESPACE else self.namespace
        return self.deploy_manager.get_object_current_state(
            kind=kind,
            name=name,
            namespace=namespace,
            api_version=api_version,
        )

    def filter_objects_current_state(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        api_version: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        namespace: Optional[str] = _SESSION_NAMESPACE,
    ) -> Tuple[bool, List[dict]]:
        """Get the current state of the given object in the namespace of this
        session

        Args:
            kind:  str
                The kind of the object to fetch
            label_selector:  str
                The label selector to filter the results by
            field_selector:  str
                The field selector to filter the results by
            api_version:  str
                The api_version of the resource kind to fetch

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  List[Dict]
                The list of resources in dict representation,
                or [] if none match
        """
        namespace = namespace if namespace != _SESSION_NAMESPACE else self.namespace
        return self.deploy_manager.filter_objects_current_state(
            kind=kind,
            namespace=namespace,
            api_version=api_version,
            label_selector=label_selector,
            field_selector=field_selector,
        )

    @alog.logged_function(log.debug2)
    @alog.timed_function(log.debug2)
    def get_status(self) -> dict:
        """Get the status of the resource being managed by this session or an
        empty dict if not available

        Returns:
            current_status:  dict
                The dict representation of the status subresource for the CR
                being managed by this session
        """

        # Pull the kind, name, and namespace
        kind = self.cr_manifest.get("kind")
        name = self.name
        api_version = self.api_version
        log.debug3("Getting status for %s.%s/%s", api_version, kind, name)

        # Fetch the current status
        success, content = self.get_object_current_state(
            kind=kind,
            name=name,
            api_version=api_version,
        )
        assert_cluster(
            success, f"Failed to fetch status for [{api_version}/{kind}/{name}]"
        )
        if content:
            return content.get("status", {})
        return {}

    ## Implementation Details ##################################################

    @staticmethod
    def _validate_cr(cr_manifest: aconfig.Config):
        """Ensure that all expected elements of the CR are present. Expected
        elements are those that are guaranteed to be present by the kube API.
        """
        assert "kind" in cr_manifest, "CR missing required section ['kind']"
        assert "apiVersion" in cr_manifest, "CR missing required section ['apiVersion']"
        assert "metadata" in cr_manifest, "CR missing required section ['metadata']"
        assert (
            "name" in cr_manifest.metadata
        ), "CR missing required section ['metadata.name']"
        assert (
            "namespace" in cr_manifest.metadata
        ), "CR missing required section ['metadata.namespace']"
